function [mean_e, std_e, density] = errorFlow(flow_1, flow_2)
% Syntax: [mean_e, std_e, density] = errorFlow(flow_1, flow_2)
%
%         flow_1        - flow matrix 1
%         flow_2        - flow matrix 2 
%         mean_e, std_3 - mean and standard deviation of the error
%         density       - density of the flow
%
% Description: Compute the flow error.
%
% Author: Haiying Liu
%   Date: August 23, 2001
%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

dbstop if error;

%===========================================================================

% Compute Error.
nRow_1   = size(flow_1, 1);
nCol_1   = size(flow_1, 2);

nRow_2   = size(flow_2, 1);
nCol_2   = size(flow_2, 2);

if nRow_1 ~= nRow_2 | nCol_1 ~= nCol_2

  disp('    WARNING: The sizes of flow_1 and flow_2 are different.');
  disp('             They are trimmed to be equal.');
  
  diffRow = abs(nRow_1 - nRow_2);
  diffCol = abs(nCol_1 - nCol_2);

  if floor(diffRow ./ 2) .* 2 ~= diffRow | ...
      floor(diffCol ./ 2) .* 2 ~= diffCol
    error('ERROR: The sizes of flow_1 and flow_2 are wrong.');
  end
    
  halfDiffRow = diffRow ./ 2;
  halfDiffCol = diffCol ./ 2;
  
  if nRow_1 > nRow_2
    flow_1 = flow_1(1 + halfDiffRow:nRow_1 - halfDiffRow, :, :);
  elseif nRow_1 < nRow_2
    flow_2 = flow_2(1 + halfDiffRow:nRow_2 - halfDiffRow, :, :);
  end
  
  if nCol_1 > nCol_2
    flow_1 = flow_1(:, 1 + halfDiffCol:nCol_1 - halfDiffCol, :);
  elseif nCol_1 < nCol_2
    flow_2 = flow_2(:, 1 + halfDiffCol:nCol_2 - halfDiffCol, :);
  end
  
end

nRow     = size(flow_1, 1);
nCol     = size(flow_1, 2);

margin   = 0;
region   = [1+margin, nRow-margin, 1+margin, nCol-margin];

rowStart = region(1);
rowEnd   = region(2);
colStart = region(3);
colEnd   = region(4);

nPoints  = (rowEnd - rowStart + 1) .* (colEnd - colStart + 1);

rows = rowStart:rowEnd;
cols = colStart:colEnd;

Vc = cat(3, flow_1(rows, cols, :), ones(length(rows), length(cols)));
Vc_normalized = Vc ./ repmat(sqrt(sum(Vc .^ 2, 3)), [1 1 3]);

Ve = cat(3, flow_2(rows, cols, :), ones(length(rows), length(cols)));
Ve_normalized = Ve ./ repmat(sqrt(sum(Ve .^ 2, 3)), [1 1 3]);

cosphi = sum(Vc_normalized .* Ve_normalized, 3);
phi    = acos(cosphi) * 180 / pi ;

mean_e = mean(phi(:));
std_e  =  std(phi(:));

density = nPoints ./ (nRow .* nCol) * 100;
