function varargout = filter_ndim(f, operator, shape)
%filter_ndim Approximate gradient.
%   [FX,FY] = FILTER_NDIM(F, OPERATOR, SHAPE) returns the numerical gradient
%   of the matrix F. FX corresponds to dF/dx, the differences in the
%   x (column) direction. FY corresponds to dF/dy, the differences
%   in the y (row) direction. The spacing between points in each
%   direction is assumed to be one. When F is a vector, DF = GRADIENT(F)
%   is the 1-D gradient.
%
%   [FX,FY,FZ] = filter_ndim(F, OPERATOR, SHAPE), when F is a 3-D array,
%   returns the numerical gradient of F. FZ corresponds to dF/dz, the
%   differences in the z direction.
%
%   [FX,FY,FZ,...] = filter_ndim(F, OPERATOR, SHAPE) extends similarly when
%   F is N-D and must be invoked with N outputs and either 2 or N+1 inputs.
%
%   Default operator = [-3 32 -168 672 0 -672 168 -32 3]' ./ 840;
%
%   Examples:
%       z       = x .* exp(-x.^2 - y.^2);
%       op      = getOperator % some operator
%       [px,py] = filter_ndim(z, op, 'same');
%
%   Haiying Liu, July 25, 2001
%   Reference: GRADIENT.M

% Loop over each dimension. Permute so that the gradient is always taken along
% the columns.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

msg = nargchk(2, 3, nargin);
if (~isempty(msg))
  error(strcat('ERROR:', msg));
end

clear msg;

if nargin < 3
  shape = 'same';
end

% if nargin < 2
%   operator = [3 -32 168 -672 0 672 -168 32 -3]' ./ 840;
% end

%==============================================================================

if size(operator, 1) ~= 1 & size(operator, 2) ~= 1
  error('ERROR: <operator> must be a vector.');
end

if size(operator, 1) == 1
  operator = operator';
end

%===============================================================================

ndim = ndims(f);
cflag = 0;
if ndims(f) == 2
   if size(f,2) == 1
      ndim  = 1;
      cflag = 1;
   elseif size(f,1) == 1    % Treat row vector as a column vector
      ndim  = 1;
      cflag = 0;
      f     = f.';
   end
end

%===============================================================================

if ndim == 1
  perm = [1, 2];
else
  perm = [2:ndim, 1]; % Cyclic permutation
end

for k = 1:ndim
   temp      = filter2(operator, f(:, :), shape);
   size_g    = size(f);
   size_g(1) = size(temp, 1);
   g         = zeros(size_g);
   g(:, :)   = temp;
   
   varargout{k} = ipermute(g,[k:ndim, 1:k-1]);

   % Set up for next pass through the loop
   f = permute(f, perm);
end 

% Swap 1 and 2 since x is the second dimension and y is the first.
if ndim > 1
  tmp = varargout{1};
  varargout{1} = varargout{2};
  varargout{2} = tmp;
end

if cflag
  varargout{1} = varargout{1}.'; 
end
