function [angleX, angleY, vDefined] = findAffineFlow(sequenceName, frame_idx)
%
% Syntax: [angleX, angleY] = findAffineFlow(sequenceName, frame_idx)
%
%         sequenceName    - the file name of sequence
%         frame_idx       - the frame index on which the flow will be calculated
%         frame_step      - index step for the sequence
%         nDigit          - 'nDigit = 4' means the '13'th frame is labeled '0013'.
%                           'nDigit = 0' means no fixed digits is defined.
%         derivativeFilterRadius
%                         - first order derivative filter radius,
%                           the filter support is 2 * radius + 1
%         correctFlowName - the file name of correct flow (*.bin)
%                           for example, frame_step = 2: 1,3,5,...
%         symbol          - symbol character used to mark indices
%         imageFormat     - image file format
%
%         For example, compute yosemite_009.jpg:
%
%           main('yosemite', 9, 1, 3, 6, 'yosemite_9_flow', '_', 'jpg')
%
% Description: Optical flow by linear structure tensor
%              (generalized eigenvalue problem).
%
% Author: Haiying Liu
%   Date: July 20, 2001
% Modified for space time segmentation: Daniel DeMenthon, August 11, 2002
%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

dbstop if error;

tic;

frame_step = 1;
nDigit = 5;
derivativeFilterRadius = 6;
symbol = '_';
imageFormat = 'bmp';

%===============================================================================
% Define parameters that tweak the algorithm

% Parameters regarding threshold
persistent THRESHOLD_LARGE_NUM;
persistent THRESHOLD_ZERO;
THRESHOLD_LARGE_NUM  = 1e6;
THRESHOLD_ZERO       = 1e-9;

% Parameters affecting accuracy/speed
persistent N_ITERATION;
persistent THRESHOLD_TRACE_STRUCTURE_TENSOR;
persistent THRESHOLD_CONFIDENCE;
persistent INIT_SMOOTH_TENSOR_SIGMA;
persistent STEP_SMOOTH_TENSOR_SIGMA;
persistent MAX_TENSOR_WIN_SIZE;
persistent SMOOTH_TENSOR_WIN_SIZE_RATE;
N_ITERATION                      = 5;
% N_ITERATION                      = 20;
THRESHOLD_TRACE_STRUCTURE_TENSOR = 1;  % 
THRESHOLD_CONFIDENCE             = 0.01;  % partly fixed
% THRESHOLD_CONFIDENCE             = 0.0;  % partly fixed
INIT_SMOOTH_TENSOR_SIGMA         = 5;
STEP_SMOOTH_TENSOR_SIGMA         = 6;
MAX_TENSOR_WIN_SIZE              = 21; % fixed. not much influence when larger than 31
SMOOTH_TENSOR_WIN_SIZE_RATE      = 4;  % winSize = SMOOTH_TENSOR_WIN_SIGMA * ...
%                                        SMOOTH_TENSOR_WIN_SIZE_RATE + 1

% Parameters regarding preproccessing
persistent DO_SMOOTH_IMAGE;
persistent SMOOTH_IMAGE_WIN_SIZE;
persistent SMOOTH_IMAGE_WIN_SIGMA;
DO_SMOOTH_IMAGE        = 1;
SMOOTH_IMAGE_WIN_SIGMA = 1;
SMOOTH_IMAGE_WIN_SIZE  = SMOOTH_IMAGE_WIN_SIGMA * 4 + 1;

% Parameters restricting computing region
persistent COMPUTE_REGION;
persistent ROW_STEP;
persistent COL_STEP;
COMPUTE_REGION = []; % '[]' means compute the whole image
ROW_STEP = 1;
COL_STEP = 1;

% Other parameters
persistent G_RATIO;
G_RATIO = 1;

persistent DO_ADD_NOISE;
persistent NOISE_MEAN;
persistent NOISE_VARIANCE;
DO_ADD_NOISE   = 0;
NOISE_MEAN     = 0;
NOISE_VARIANCE = 0.05;

persistent MAX_FLOW_MAGNITUDE;
MAX_FLOW_MAGNITUDE = 5;

persistent RESULT_IMAGE_FORMAT;
RESULT_IMAGE_FORMAT = 'eps';

persistent DO_HAVE_WARNED
DO_HAVE_WARNED = 0;

persistent IMAGE_SMOOTH_FILTER_SHAPE;
persistent GRADIENT_FILTER_SHAPE;
persistent TENSOR_SMOOTH_FILTER_SHAPE;
IMAGE_SMOOTH_FILTER_SHAPE  = 'same';
GRADIENT_FILTER_SHAPE      = 'valid';
TENSOR_SMOOTH_FILTER_SHAPE = 'same';

persistent FILE_DIR;
persistent RESULT_DIR;
% FILE_DIR   = '../../Data';
% RESULT_DIR = './Result';
FILE_DIR   = '';
RESULT_DIR = '.';

%===============================================================================
%= Prepare directory.

filepath     = [sequenceName];

%===============================================================================
%= Read/pre-process the image sequence

disp('Reading the image sequence...');

operator = firstOrderDerivative(derivativeFilterRadius);

frameSpan = floor(length(operator) ./ 2);

beginIdx  = frame_idx - frameSpan * frame_step;
endIdx    = frame_idx + frameSpan * frame_step;

nFrame    = (endIdx - beginIdx) / frame_step + 1;
disp(['  Number of frames: ', num2str(nFrame)]);

if DO_SMOOTH_IMAGE
  filter_smooth = fspecial('gaussian', ...
    SMOOTH_IMAGE_WIN_SIZE, SMOOTH_IMAGE_WIN_SIGMA);
end

for idx = beginIdx:frame_step:endIdx
  
  idxStr    = num2str(idx);
  if nDigit > 0
    while length(idxStr) < nDigit
      idxStr  = ['0', idxStr];
    end
  end
  
%  imageName = [filepath, symbol, idxStr, '.', imageFormat];
  imageName = [filepath, idxStr, '.', imageFormat];
  rgbImage = imread(imageName);
  thisImage =  double(rgb2gray(rgbImage));
  
  if DO_ADD_NOISE
    thisImage = ...
      imnoise(thisImage, 'gaussian', NOISE_MEAN, NOISE_VARIANCE);
  end
  
  if DO_SMOOTH_IMAGE
    thisImage = filter2(filter_smooth, thisImage, IMAGE_SMOOTH_FILTER_SHAPE);
  end
  
  thisIdx   = (idx - beginIdx) / frame_step + 1;
  sequence(:, :, thisIdx) = thisImage;
  
end

%===============================================================================
%= Read correct flow

%if exist(correctFlowName, 'file')
%  correctFlow = readCorrectFlow(correctFlowName);
%else
%  disp('WARNING: Cannot find correct flow file.');
%  disp('         Thus there is no error statistics.');
%end

%===============================================================================
%= Compute the gradient

disp('Computing the gradient gx, gy, gt...');

[gx, gy, gt] = filter_ndim(sequence, operator, GRADIENT_FILTER_SHAPE);

% Get gradient of current frame
thisFrame = (frame_idx - beginIdx) / frame_step + 1;

gx = gx(:, :, thisFrame) * G_RATIO;
gy = gy(:, :, thisFrame) * G_RATIO;
gt = gt(:, :, round(size(gt, 3) ./ 2)) * G_RATIO;

% Extract the valid area
nRow = min([size(gx, 1), size(gy, 1), size(gt, 1)]);
nCol = min([size(gx, 2), size(gy, 2), size(gt, 2)]);

disp(['  Image size = ', num2str(size(sequence(:, :, 1))), ' (row col)']);
disp(['  Valid size = ', num2str([nRow nCol]), ' (row col)']);

gx = wkeep(gx, [nRow, nCol]);
gy = wkeep(gy, [nRow, nCol]);
gt = wkeep(gt, [nRow, nCol]);

nFrame = size(sequence , 3);

if isempty(COMPUTE_REGION)
  COMPUTE_REGION = [1, nRow, 1, nCol];
end

%===============================================================================
%= Compute structure tensor components

disp('Computing structure tensor components...');

% G_xx, G_xy, G_yy, G_xt, G_yt, G_tt

switch size(gt, 3)
case 1
  gt = gt * G_RATIO;
case nFrame
  gt = gt(:,:,thisFrame) * G_RATIO;
otherwise
  disp('WARNING: Dimension of "gt" might be wrong.');
  pause;
end

% Compose elements of the structure tensor
[x, y] = meshgrid(1:nCol, 1:nRow);

nSize_tensor = 9;

g_11 = x .* x .* gx .* gx;
g_12 = x .* y .* gx .* gx;
g_13 = x .* x .* gx .* gy;
g_14 = x .* y .* gx .* gy;
g_15 = x .* x .* gx .* gt;
g_16 = x .* y .* gx .* gt;
g_17 = x .*      gx .* gx;
g_18 = x .*      gx .* gy;
g_19 = x .*      gx .* gt;

g_22 = y .* y .* gx .* gx;
g_23 = x .* y .* gx .* gy;
g_24 = y .* y .* gx .* gy;
g_25 = x .* y .* gx .* gt;
g_26 = y .* y .* gx .* gt;
g_27 = y .*      gx .* gx;
g_28 = y .*      gx .* gy;
g_29 = y .*      gx .* gt;

g_33 = x .* x .* gy .* gy;
g_34 = x .* y .* gy .* gy;
g_35 = x .* x .* gy .* gt;
g_36 = x .* y .* gy .* gt;
g_37 = x      .* gx .* gy;
g_38 = x      .* gy .* gy;
g_39 = x      .* gy .* gt;

g_44 = y .* y .* gy .* gy;
g_45 = x .* y .* gy .* gt;
g_46 = y .* y .* gy .* gt;
g_47 = y      .* gx .* gy;
g_48 = y      .* gy .* gy;
g_49 = y      .* gy .* gt;

g_55 = x .* x .* gt .* gt;
g_56 = x .* y .* gt .* gt;
g_57 = x .*      gx .* gt;
g_58 = x .*      gy .* gt;
g_59 = x .*      gt .* gt;

g_66 = y .* y .* gt .* gt;
g_67 = y .*      gx .* gt;
g_68 = y .*      gy .* gt;
g_69 = y .*      gt .* gt;

g_77 =           gx .* gx;
g_78 =           gx .* gy;
g_79 =           gx .* gt;

g_88 =           gy .* gy;
g_89 =           gy .* gt;

g_99 =           gt .* gt;

% Initialize structure tensor components

G_11 = g_11;
G_12 = g_12;
G_13 = g_13;
G_14 = g_14;
G_15 = g_15;
G_16 = g_16;
G_17 = g_17;
G_18 = g_18;
G_19 = g_19;

G_22 = g_22;
G_23 = g_23;
G_24 = g_24;
G_25 = g_25;
G_26 = g_26;
G_27 = g_27;
G_28 = g_28;
G_29 = g_29;

G_33 = g_33;
G_34 = g_34;
G_35 = g_35;
G_36 = g_36;
G_37 = g_37;
G_38 = g_38;
G_39 = g_39;

G_44 = g_44;
G_45 = g_45;
G_46 = g_46;
G_47 = g_47;
G_48 = g_48;
G_49 = g_49;

G_55 = g_55;
G_56 = g_56;
G_57 = g_57;
G_58 = g_58;
G_59 = g_59;

G_66 = g_66;
G_67 = g_67;
G_68 = g_68;
G_69 = g_69;

G_77 = g_77;
G_78 = g_78;
G_79 = g_79;

G_88 = g_88;
G_89 = g_89;

G_99 = g_99;

%================================================================================
%= Structure tensor analysis

disp('Analysing structure tensor...');

% Initializate flow etc.
tensor   = zeros(9);

flow     = zeros(nRow, nCol, 2);
lastFlow = zeros(nRow, nCol, 2);

c_t      = zeros(nRow, nCol);
c_s      = zeros(nRow, nCol);
c_c      = zeros(nRow, nCol);
last_c_t = zeros(nRow, nCol);
last_c_s = zeros(nRow, nCol);
last_c_c = zeros(nRow, nCol);
c_cMean  = ones(N_ITERATION, 1) .* NaN;
c_cStd   = ones(N_ITERATION, 1) .* NaN;

% Initialize error measurement
errorMean = ones(N_ITERATION, 1) .* Inf;
errorStd  = ones(N_ITERATION, 1) .* Inf;

% Initialize confidence measurement
[x, y] = meshgrid(COMPUTE_REGION(3):COL_STEP:COMPUTE_REGION(4), ...
  COMPUTE_REGION(1):ROW_STEP:COMPUTE_REGION(2));
pixels = [x(:), y(:)];
lowConfidencePixelFlag = ones(length(pixels), 1);

% Set gaussian sigma for window function
sigma_total = INIT_SMOOTH_TENSOR_SIGMA;

% Begin estimating optical flow
for iteration = 1:N_ITERATION
  
  % Check if there are any low confidence pixels
  nLowConfidencePixels = sum(lowConfidencePixelFlag);
  if nLowConfidencePixels == 0
    break;
  end
  
  % Define window function for tensor smoothing
  if iteration > 1
    sigma_cascade = sqrt( ...
      (sigma_total + STEP_SMOOTH_TENSOR_SIGMA) * ...
      (sigma_total + STEP_SMOOTH_TENSOR_SIGMA) - ...
      (sigma_total * sigma_total) );
    sigma_total = sigma_total + STEP_SMOOTH_TENSOR_SIGMA;
  else
    sigma_cascade = sigma_total;
  end
  
  winSize = sigma_cascade * SMOOTH_TENSOR_WIN_SIZE_RATE + 1;
  if winSize > MAX_TENSOR_WIN_SIZE
    winSize = MAX_TENSOR_WIN_SIZE;
  end
  
  window  = fspecial('gaussian', winSize, sigma_cascade);
  
  % Display interation information
  disp(['# iteration = ', num2str(iteration), ...
      ' (', datestr(now, 13), ') ---------']);
  
  disp(['    low confidence pixels: ', ...
      num2str(nLowConfidencePixels)]);
  
  disp(['    sigma_step    = ', num2str(STEP_SMOOTH_TENSOR_SIGMA)]);
  disp(['    sigma_cascade = ', num2str(sigma_cascade)]);
  disp(['    sigma_total   = ', num2str(sigma_total)]);
  
  % Compute the average of the element of the structure tensor
  % over its neighbor
  smoothFilterShape = 'same';
  
  G_11 = filter2(window, G_11, TENSOR_SMOOTH_FILTER_SHAPE);
  G_12 = filter2(window, G_12, TENSOR_SMOOTH_FILTER_SHAPE);
  G_13 = filter2(window, G_13, TENSOR_SMOOTH_FILTER_SHAPE);
  G_14 = filter2(window, G_14, TENSOR_SMOOTH_FILTER_SHAPE);
  G_15 = filter2(window, G_15, TENSOR_SMOOTH_FILTER_SHAPE);
  G_16 = filter2(window, G_16, TENSOR_SMOOTH_FILTER_SHAPE);
  G_17 = filter2(window, G_17, TENSOR_SMOOTH_FILTER_SHAPE);
  G_18 = filter2(window, G_18, TENSOR_SMOOTH_FILTER_SHAPE);
  G_19 = filter2(window, G_19, TENSOR_SMOOTH_FILTER_SHAPE);
  
  G_22 = filter2(window, G_22, TENSOR_SMOOTH_FILTER_SHAPE);
  G_23 = filter2(window, G_23, TENSOR_SMOOTH_FILTER_SHAPE);
  G_24 = filter2(window, G_24, TENSOR_SMOOTH_FILTER_SHAPE);
  G_25 = filter2(window, G_25, TENSOR_SMOOTH_FILTER_SHAPE);
  G_26 = filter2(window, G_26, TENSOR_SMOOTH_FILTER_SHAPE);
  G_27 = filter2(window, G_27, TENSOR_SMOOTH_FILTER_SHAPE);
  G_28 = filter2(window, G_28, TENSOR_SMOOTH_FILTER_SHAPE);
  G_29 = filter2(window, G_29, TENSOR_SMOOTH_FILTER_SHAPE);
  
  G_33 = filter2(window, G_33, TENSOR_SMOOTH_FILTER_SHAPE);
  G_34 = filter2(window, G_34, TENSOR_SMOOTH_FILTER_SHAPE);
  G_35 = filter2(window, G_35, TENSOR_SMOOTH_FILTER_SHAPE);
  G_36 = filter2(window, G_36, TENSOR_SMOOTH_FILTER_SHAPE);
  G_37 = filter2(window, G_37, TENSOR_SMOOTH_FILTER_SHAPE);
  G_38 = filter2(window, G_38, TENSOR_SMOOTH_FILTER_SHAPE);
  G_39 = filter2(window, G_39, TENSOR_SMOOTH_FILTER_SHAPE);
  
  G_44 = filter2(window, G_44, TENSOR_SMOOTH_FILTER_SHAPE);
  G_45 = filter2(window, G_45, TENSOR_SMOOTH_FILTER_SHAPE);
  G_46 = filter2(window, G_46, TENSOR_SMOOTH_FILTER_SHAPE);
  G_47 = filter2(window, G_47, TENSOR_SMOOTH_FILTER_SHAPE);
  G_48 = filter2(window, G_48, TENSOR_SMOOTH_FILTER_SHAPE);
  G_49 = filter2(window, G_49, TENSOR_SMOOTH_FILTER_SHAPE);
  
  G_55 = filter2(window, G_55, TENSOR_SMOOTH_FILTER_SHAPE);
  G_56 = filter2(window, G_56, TENSOR_SMOOTH_FILTER_SHAPE);
  G_57 = filter2(window, G_57, TENSOR_SMOOTH_FILTER_SHAPE);
  G_58 = filter2(window, G_58, TENSOR_SMOOTH_FILTER_SHAPE);
  G_59 = filter2(window, G_59, TENSOR_SMOOTH_FILTER_SHAPE);
  
  G_66 = filter2(window, G_66, TENSOR_SMOOTH_FILTER_SHAPE);
  G_67 = filter2(window, G_67, TENSOR_SMOOTH_FILTER_SHAPE);
  G_68 = filter2(window, G_68, TENSOR_SMOOTH_FILTER_SHAPE);
  G_69 = filter2(window, G_69, TENSOR_SMOOTH_FILTER_SHAPE);
  
  G_77 = filter2(window, G_77, TENSOR_SMOOTH_FILTER_SHAPE);
  G_78 = filter2(window, G_78, TENSOR_SMOOTH_FILTER_SHAPE);
  G_79 = filter2(window, G_79, TENSOR_SMOOTH_FILTER_SHAPE);
  
  G_88 = filter2(window, G_88, TENSOR_SMOOTH_FILTER_SHAPE);
  G_89 = filter2(window, G_89, TENSOR_SMOOTH_FILTER_SHAPE);
  
  G_99 = filter2(window, G_99, TENSOR_SMOOTH_FILTER_SHAPE);
  
  % Compose tensors.
  J = zeros(nSize_tensor, nSize_tensor, nRow .* nCol);
  
  J(1, 2, :) = G_12(:);
  J(1, 3, :) = G_13(:);
  J(1, 4, :) = G_14(:);
  J(1, 5, :) = G_15(:);
  J(1, 6, :) = G_16(:);
  J(1, 7, :) = G_17(:);
  J(1, 8, :) = G_18(:);
  J(1, 9, :) = G_19(:);
  
  J(2, 3, :) = G_23(:);
  J(2, 4, :) = G_24(:);
  J(2, 5, :) = G_25(:);
  J(2, 6, :) = G_26(:);
  J(2, 7, :) = G_27(:);
  J(2, 8, :) = G_28(:);
  J(2, 9, :) = G_29(:);
  
  J(3, 4, :) = G_34(:);
  J(3, 5, :) = G_35(:);
  J(3, 6, :) = G_36(:);
  J(3, 7, :) = G_37(:);
  J(3, 8, :) = G_38(:);
  J(3, 9, :) = G_39(:);
  
  J(4, 5, :) = G_45(:);
  J(4, 6, :) = G_46(:);
  J(4, 7, :) = G_47(:);
  J(4, 8, :) = G_48(:);
  J(4, 9, :) = G_49(:);
  
  J(5, 6, :) = G_56(:);
  J(5, 7, :) = G_57(:);
  J(5, 8, :) = G_58(:);
  J(5, 9, :) = G_59(:);
  
  J(6, 7, :) = G_67(:);
  J(6, 8, :) = G_68(:);
  J(6, 9, :) = G_69(:);
  
  J(7, 8, :) = G_78(:);
  J(7, 9, :) = G_79(:);
  
  J(8, 9, :) = G_89(:);
  
  tensor_total = zeros(nSize_tensor, nSize_tensor, nRow .* nCol);
  tensor_total(1, 1, :) = G_11(:);
  tensor_total(2, 2, :) = G_22(:);
  tensor_total(3, 3, :) = G_33(:);
  tensor_total(4, 4, :) = G_44(:);
  tensor_total(5, 5, :) = G_55(:);
  tensor_total(6, 6, :) = G_66(:);
  tensor_total(7, 7, :) = G_77(:);
  tensor_total(8, 8, :) = G_88(:);
  tensor_total(9, 9, :) = G_99(:);
  
  tensor_total = tensor_total + J + permute(J, [2, 1, 3]);
  
  % Begin eigenvalue analysis
  lowConfidendPixelIdx = find(lowConfidencePixelFlag);
  
  disp(['    Tensor analysis (', datestr(now, 13), ')']);
  
  for index = 1:length(lowConfidendPixelIdx)
    
    % Get low confidend pixel
    pxIdx = lowConfidendPixelIdx(index);
    
    x = pixels(pxIdx, 1);
    y = pixels(pxIdx, 2);
    
    tensor = tensor_total(:, :, pxIdx);
    
    % Compute eigenvector and eigenvalue
    B = [ ...
        x y 0 0 0 0 1 0 0; ...
        0 0 x y 0 0 0 1 0; ...
        0 0 0 0 x y 0 0 1 ];
    
    D = B' * B;
    
    [eigV, eigD]     = eig(tensor, D);
    eigD             = diag(eigD);
    
    % For debug purpose only
    rawEigV = eigV;
    rawEigD = eigD;
    
    %if x == 15 & y == 48
    %  disp('DEBUG POINT');
    %end
    
    % Extract finite eigenvalue with its eigenvector
    finiteEigD_idx   = find(isfinite(eigD));
    eigD             = eigD(finiteEigD_idx);
    eigV             = eigV(:, finiteEigD_idx);
    
    % Remove large positive eigenvalue
    normal_idx       = find(eigD < THRESHOLD_LARGE_NUM);
    eigD             = eigD(normal_idx);
    eigV             = eigV(:, normal_idx);
    
    % Remove negative eigenvalue with its eigenvector
    positiveEigD_idx = find(eigD > -THRESHOLD_ZERO);
    eigD             = eigD(positiveEigD_idx);
    eigV             = eigV(:, positiveEigD_idx);
    
    % Sort eigenvalue in ascending order
    % i.e. eigD(1) <= eigD(2) <= eigD(3)
    [eigD, sort_idx] = sort(eigD);
    eigV             = eigV(:, sort_idx);
    
    % Get the smallest 3
    if length(eigD) > 3 & ~DO_HAVE_WARNED
      disp('WARNING: The valid eigenvalues are more than 3.');
      disp('         Pick 3 smallest.');
      disp('rawEigD =');
      disp(num2str(rawEigD));
      disp('eigD =');
      disp(num2str(eigD));
      
      eigD = eigD(1:3);
      eigV = eigV(:, 1:3);
      
      DO_HAVE_WARNED = 1;
    end
    
    % Should not have complex eigenvalue,
    % since both 'tensor' and 'D' are symmetric
    imagPart = imag(eigD);
    if sum(abs(imagPart)) ~= 0
      disp('WARNING: None-zero image part.');
      continue;
    end
    
    % Trace
    validTraceTensor = sum(eigD);
    
    % Eigenvalue/eigenvector analysis
    
    if validTraceTensor >= THRESHOLD_TRACE_STRUCTURE_TENSOR
      
      % Initalize coherency
      coherency_total   = 0;
      coherency_spatial = 0;
      coherency_corner  = 0;
      
      % Compute total, spatial, and corner coherency
      if length(eigD) == 3
        temp = (eigD(3) - eigD(1)) ./ (eigD(3) + eigD(1));
        coherency_total   = temp .* temp;
        
        temp = (eigD(3) - eigD(2)) ./ (eigD(3) + eigD(2));
        coherency_spatial = temp .* temp;
        
        coherency_corner  = coherency_total - coherency_spatial;
      elseif ~DO_HAVE_WARNED
        % For debug only.
        disp('WARNING: The number of eigenvalue is incorrect,');
        disp('         possibly due to the numerical error.');
        
        DO_HAVE_WARNED = 1;
      end
      
      % Compute the flow only if corner coherency is increased
      if coherency_corner < c_c(y, x)
        if c_c(y, x) > THRESHOLD_CONFIDENCE
          lowConfidencePixelFlag(pxIdx) = 0;
        end
        
        continue;
      end
      
      % Compute flow or normal flow, depending on the confidence
      if coherency_corner > THRESHOLD_CONFIDENCE | length(eigD) ~= 3
        % Optical flow
        v1 = B * eigV(:, 1);
        if v1(3) ~= 0
          thisFlow = [v1(1), v1(2)] ./ v1(3);
        end
      else
        % Normal flow
        v3 = B * eigV(:, 3);
        thisFlow = -[v3(1), v3(2)] .* v3(3) ./ ...
          (v3(1) .* v3(1) + v3(2) .* v3(2));
      end
      
      % Remove outliers
      if abs(thisFlow) <= MAX_FLOW_MAGNITUDE
        % Save coherency
        c_t(y, x)  = coherency_total;
        c_s(y, x)  = coherency_spatial;
        c_c(y, x)  = coherency_corner; % used as confidence
        
        flow(y, x, 1) = thisFlow(1);
        flow(y, x, 2) = thisFlow(2);
      end
      
    end % if trace_tensor >= THRESHOLD_TRACE_STRUCTURE_TENSOR
    
  end % index
  
  %===============================================================================

%= Compute error
  
% if exist('correctFlow', 'var') & ~isempty(correctFlow)
  if 0
    
    [errorMean(iteration), errorStd(iteration), density] = ...
      error_flow(flow, correctFlow);
    
    disp(['    mean confidence = ', num2str(mean(c_c(:)))]);
    disp(['    error mean      = ', num2str(errorMean(iteration))]);
    disp(['    error std       = ', num2str(errorStd(iteration))]);
    disp(['    density         = ', num2str(density), '%']);
    
    if iteration > 1
      if errorMean(iteration) > errorMean(iteration - 1)
        flow = lastFlow;
        c_t  = last_c_t;
        c_s  = last_c_s;
        c_c  = last_c_c;
        disp('    ### Use result from last iteration. ###');
        %        break;
      end
    end
    
  else
    
    disp(['    mean confidence = ', num2str(mean(c_c(:)))]);
    
    if mean(c_c(:)) < mean(last_c_c(:))
      flow = lastFlow;
      c_t  = last_c_t;
      c_s  = last_c_s;
      c_c  = last_c_c;
      disp('    ### Use result from last iteration. ###');
      break;
    end
    
  end
  
  lastFlow = flow;
  last_c_t = c_t;
  last_c_s = c_s;
  last_c_c = c_c;
  
  c_cMean(iteration) = mean(c_c(:));
  c_cStd(iteration)  = std(c_c(:));
  
end % iteration

angleX = uint8(90 - atan(flow(:,:,1)) * 180/pi);
angleY = uint8(90 - atan(flow(:,:,2)) * 180/pi);

vDefined = uint8(logical(c_c > THRESHOLD_CONFIDENCE));
toc;
