function operator = firstOrderDerivative(opRadius)
% Syntax: operator = firstOrderDerivative(opRadius)
%
%         opRadius - operator radius
%         operator - first-order derivative operator
%                    whose size is [1, opRadius * 2 + 1].
%
% Description: Generate series-designed first-order derivative operator.
%
% Reference:
% @Book{Haussecker_1,
% author  = "H. Haussecker and H. Spies",
% title   = "Handbook of Computer Vision and Applications Volume 2: Signal Processing and Pattern Recognition, Chapter 13: Motion",
% publisher = "Academic Press",
% year    = 1999",
% pages   = "309-396",
%
% Author: Haiying Liu
%   Date: Sept. 19, 2001
%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

dbstop if error;

tic;

msg = nargchk(1, 1, nargin);
if (~isempty(msg))
  error(strcat('ERROR:', msg));
end

clear msg;

%===============================================================================
% The operator is composed by solving the linear system below:
%  _                                       _   _     _     _   _
% |   1    2        3        ...  R         | |  d_1  |   |  1  |
% |   1    8        27       ...  R^3       | |  d_2  |   |  0  |
% |   1    32       243      ...  R^5       | |  d_3  | = |  0  |
% |   ...  ...      ...      ...  ...       | |  ...  |   | ... |
% |_  1    2^(2R-1) 3^(2R-1) ...  R^(2R-1) _| |_ d_R _|   |_ 0 _|
%                    ||                          ||         ||
%                     A                           x     =    b

A(1,:) = 1:opRadius;
b     = [1; zeros(opRadius - 1, 1)];

for row = 2:opRadius
  A(row, :) = A(row - 1, :) .* A(1, :) .* A(1, :);
end

x = A \ b;

operator = [-fliplr(x'), 0, x'] ./ 2;