function [match, globalDist] = findHungarianSequenceDistance(list1, list2, nbFrames1, nbFrames2, importances, distThreshold)

% Usage: [match, globalDist] = findHungarianSequenceDistance(list1, list2, importances, timeShift)

% Find the minimum distance between 2 lists of feature vectors that have to be matched
% Match is list of what column number matches each row
% so it is the trace of list1
% Check issues of what happens when list2 is longer than list1: mergeTraces may not work
% from 2 consecutive sequences 1 and 2
% Distance here is average distance per matched pair
% timeShift here is given, and is equal half the sum of nbFrames of sequence 1 and sequence 2
% When the timeShift is not known, use findLUVSequenceDistance

% We match feature points of list2 to points of list1, so list2 are on rows and list1 on column
% Then we will give list2 the labels of list1: label2(i) = labels1(match(i));

% Example 
% list1 = [  60.0000   90.0000    2.5000         0         0         0         0; ...
%  			120.0000   90.0000   32.1410         0         0         0         0; ...
%   		 90.0000   90.0000         0         0         0         0         0; ...
%   		 90.0000   90.0000   30.0000   30.0000   10.0000   10.0000   10.0000];
% 			
% list 2 = [ 60.0000   90.0000    8.6867         0         0         0         0; ...
%  		    120.0000   90.0000   25.9541         0         0         0         0; ...
%   		90.0000   90.0000         0         0         0         0         0];
% importances = ones[1, 7];
%
% [match, globalDist] = findHungarianSequenceDistance(list1, list2, 10, 16, importances);
%
% Outputs:
% match = 1     2     3  % 3 feature vectors ol list2 match first 3 feature vectors of list1
%
% globalDist =  0.1962
%
 			


msg = nargchk(5, 6, nargin);
if (~isempty(msg))
	error(strcat('ERROR:', msg));
end

timeShift = -(nbFrames1 + nbFrames2)/2; % negative as we had opposite convention in function below

nbPts1 = size(list1,1);
nbPts2 = size(list2,1);


distMat = zeros(nbPts1, nbPts2); % we want to match list2 to list1, not the other way around, so we will transpose it

for i = 1:nbPts1
	for j = 1:nbPts2
		distMat(i, j) = findLUVFeatureDistance(list1(i,:), list2(j,:), timeShift, importances);
	end
end

if nargin == 5
	[match, globalDist] = nonSquareHungarian(distMat'); % transpose as we want to go from list2 to list1
else % distThreshold is 6th argument
	[match, globalDist] = nonSquareHungarian(distMat', distThreshold);
end


	



