function [xStill, yStill, angleStill] = findStillFeatures(uBlock, vBlock, squaredMotionThreshold)
%
% Goal is to define strand positions that are independent of time offset of block
% to allow for retrieval of offset blocks
% For strands that are vertical (zero motion), the pixel positions x and y define the strands
% while for fast motion, the projection of an origin 
% onto the projection of the strand in a plane (t=0,x,y)
% is independent of time
% But this distance is not defined for zero motion
% so we define a position that evolves from (x, y) when motion is zero
% to the projection of an origin  when motion is larger than a threshold
% 
% Computation of projection of origin on plane parallel to time axis containing strand: 
% 1. Find unit normal n to time vector (1,0,0) and to motion vector (1,u,v). 
% 2. Find lambda = [x, y] * n
% 3. [xStill, yStill] = lambda * n
%
% The angle is defined within the plane parallel to the time axis that contains the strand
% It is the arc-cos of the dot product of normalized [0,u,v] and [1,u,v], with sign of vx - uy
% Problem: there can be a discontinuity in angle if the motion almost passes through the origin
% It may be better to use the two angles angleX and angleY.

[nbFrames, nbRows, nbCols] = size(uBlock);
[tCountBlock, yCountBlock, xCountBlock] = makeCountBlocks(nbFrames, nbRows, nbCols);
if 0
xVect = zeros(1,1,nbCols);
xVect(1,1,:) = [1:nbCols];
xCountBlock = repmat(xVect, [nbFrames, nbRows, 1]);
yVect = zeros(1,nbRows,1);
yVect(1,:,1) = [1:nbRows];
yCountBlock = repmat(yVect, [nbFrames, 1, nbCols]);
tVect = zeros(nbFrames,1,1);
tVect(:,1,1) = [1:nbFrames];
tCountBlock = repmat(tVect, [1, nbRows, nbCols]);
end
xStill = zeros(nbFrames, nbRows, nbCols);
yStill = zeros(nbFrames, nbRows, nbCols);

squaredMotion = uBlock .* uBlock + vBlock .* vBlock;
isSmallMotion = logical(squaredMotion < squaredMotionThreshold);

uxvy = uBlock .* xCountBlock + vBlock .* yCountBlock;

xStill(isSmallMotion) = xCountBlock(isSmallMotion) - uBlock(isSmallMotion) .* uxvy(isSmallMotion) ./ squaredMotionThreshold;
yStill(isSmallMotion) = yCountBlock(isSmallMotion) - vBlock(isSmallMotion) .* uxvy(isSmallMotion) ./ squaredMotionThreshold;

isLargeMotion = ~isSmallMotion;
xStill(isLargeMotion) = xCountBlock(isLargeMotion) - uBlock(isLargeMotion) .* uxvy(isLargeMotion) ./ squaredMotion(isLargeMotion);
yStill(isLargeMotion) = yCountBlock(isLargeMotion) - vBlock(isLargeMotion) .* uxvy(isLargeMotion) ./ squaredMotion(isLargeMotion);

signvx_uy = sign(vBlock .* xStill - uBlock .* yStill);

cosAngle = signvx_uy .* sqrt(squaredMotion ./(1 + squaredMotion));
angleStill = (180/pi) .* acos(cosAngle);

