function [angleMapX, angleMapY, flowDefinedBlock] = makePyramidAngleMaps(greyBlock)

% Usage: [angleMapX, angleMapY, flowDefinedBlock] = makePyramidAngleMaps(grey3D)
% Compute optical flow using Lucas-Kanade algorithm
% Here we don't return optical flow itself, only space-time angles. 
% we compute flow at maximum scale,
% leaving holes where the texture is absent
% then we go to half scale and recompute flow in the holes
% leaving holes again,
% then bring back these results to full scale image
% and so on until there are no more holes
% We don't crop filtered images to prevent misalignment problems
% Resulting motion images can be cropped at same time as RGB images after calling this routine
% Required cropping is 10 frames because of gaussian smoothing + 4 frames because of derivation
% i.e. 14 frames, 7 frames on each side of pixel blocks 
% since we remove filter shifting in filter3D and filter1D functions

disp(' Pyramid method for computing optical flow ');

grey3D = double(greyBlock);
[nbFrames, nbRows, nbCols] = size(grey3D);
flowDefined = ones(nbFrames, nbRows, nbCols); % places where flow is defined

% Apply spatio-temporal gaussian filter
gaussF = (exp(-[-5:5].^2/(2*1.5*1.5))/sqrt(2*pi))'; % gaussian with sigma 1.5
gaussF = gaussF/sum(gaussF); % normalizing

grey3D = filter3D(grey3D, gaussF);


uBlock = zeros(nbFrames, nbRows, nbCols); % motion fields
vBlock = zeros(nbFrames, nbRows, nbCols); % motion fields
angleMapX = zeros(nbFrames, nbRows, nbCols); % motion field angles in x and y directions
angleMapY = zeros(nbFrames, nbRows, nbCols); % if 0, no motion was found
flowDefinedBlock = logical(zeros(nbFrames, nbRows, nbCols));
lambda = zeros(2, nbRows, nbCols);

d = [-1, 8, 0, -8, 1]/12; % derivative template
mask = [0.0625,0.25,0.375,0.25,0.0625]; % weight template, sums to 1

maxNbIterations = 16;
scaling = 1;
smallNbRows = nbRows; smallNbCols = nbCols;
smallGrey3D = grey3D;

for holeFillingCount = 1:maxNbIterations
	[tmpNbFrames, smallNbRows, smallNbCols] = size(smallGrey3D);
	disp(['Step for filling motion holes # ', num2str(holeFillingCount)]);
     
	It = filter1D(smallGrey3D,d,1); 
	Iy = filter1D(smallGrey3D,d,2);
	Ix = filter1D(smallGrey3D,d,3);

	IxIx=Ix.*Ix;
	IyIy=Iy.*Iy;
	IxIy=Ix.*Iy;
	IxIt=Ix.*It;
	IyIt=Iy.*It;

	for i =1:nbFrames
        disp(i);
        clear fIxIx fIyIy fIxIy fIxIt fIyIt lambda prevFlowDefined;
        fIxIx(:,:) = IxIx(i,:,:); fIyIy(:,:) = IyIy(i,:,:); fIxIy(:,:) = IxIy(i,:,:);
        fIxIt(:,:) = IxIt(i,:,:); fIyIt(:,:) = IyIt(i,:,:);
        fIxIx=conv2(mask, mask, fIxIx, 'same');	
        fIyIy=conv2(mask, mask, fIyIy, 'same');	
        fIxIy=conv2(mask, mask, fIxIy, 'same');	
        fIxIt=conv2(mask, mask, fIxIt, 'same');	
        fIyIt=conv2(mask, mask, fIyIt, 'same');	
	
        det = fIxIx .* fIyIy - fIxIy .* fIxIy;
        rhsU = -fIxIt .* fIyIy + fIyIt .* fIxIy;
        rhsV = fIxIt .* fIxIy - fIyIt .* fIxIx;
        
		diff = fIxIx - fIyIy;
		tmp =(fIxIx + fIyIy)*0.5;
		disc = 0.5 * sqrt(diff.^2 + 4.*(fIxIy.^2));
		lambda(1,:,:) = tmp - disc;
        lambda(2,:,:) = tmp + disc;
	
        minLambda = squeeze(min(lambda,[], 1)); % can lambda be negative? check: no
		
		prevFlowDefined(:,:) = flowDefinedBlock(i,:,:);
        smallFlowDefined = logical(minLambda > 0.1); % threshold suggested by Barron in IJCV paper is 1
        prevU(:,:) = uBlock(i,:,:);
        prevV(:,:) = vBlock(i,:,:);
        smallU = zeros(smallNbRows, smallNbCols); 
        smallV = zeros(smallNbRows, smallNbCols);
        smallU(smallFlowDefined) = rhsU(smallFlowDefined) ./ det(smallFlowDefined);
        smallV(smallFlowDefined) = rhsV(smallFlowDefined) ./ det(smallFlowDefined);
		u = (nbCols/smallNbCols) * imresize(smallU, [nbRows,nbCols],'bicubic'); % back to full scale;
		v = (nbRows/smallNbRows) * imresize(smallV, [nbRows,nbCols],'bicubic'); % zero for ~flowDefined
		flowDefined = logical(imresize(smallFlowDefined, [nbRows,nbCols],'bicubic'));
        if i == 1
            disp('debug');
        end;
		u(prevFlowDefined) = 0; v(prevFlowDefined) = 0; % do not update previous calculations
        uBlock(i,:,:) = prevU + u; % u zero where prevU defined, prevU zero where defined
        vBlock(i,:,:) = prevV + v;
		flowDefinedBlock(i,:,:) = flowDefined;
	end; % for i
	
	flowDefinedMinus = flowDefinedBlock(3:nbFrames-2,3:nbRows-2,3:nbCols-2); % remove edges first
	numberHoles = sum(~flowDefinedMinus(:));
	disp(['Nb of holes is ', int2str(numberHoles)]);
	if numberHoles == 0
		break; % get out of while loop
    end; % if
    
    scaling = scaling/2; % divide grey3D by 2 in row and col directions

    smallGrey3D = blockResize(grey3D, scaling); % note: no reduction in nbFrames
    [tmpNbFrames, smallNbRows, smallNbCols] = size(smallGrey3D); % used in setting smallU, smallV
end; % for holeFillingCount

angleMapX = uint8(90 - atan(uBlock) * 180/pi);
angleMapY = uint8(90 - atan(vBlock) * 180/pi);
trim = 4; % nb of pixels to trim along each edge
angleMapX = angleMapX(trim+1:nbFrames-trim,trim+1:nbRows-trim,trim+1:nbCols-trim);
angleMapY = angleMapY(trim+1:nbFrames-trim,trim+1:nbRows-trim,trim+1:nbCols-trim);
flowDefinedBlock = flowDefinedBlock(trim+1:nbFrames-trim,trim+1:nbRows-trim,trim+1:nbCols-trim);
