%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function gradient = my_gradient(data)
% Syntax: gradient =  my_gradient(data)
%
%         data        - data to be computed
%         gradient    - a structure contains results:
%                         .mag  : magnitude of the gradient
%                                 larger than 'threshold'
%                         .dir  : direction of the gradient
%                         .mask : mask for local maxima
%                                 1 : is local maxima
%                                 0 : is not local maxima
%                       .mag .* .mask yields edges.
%
% Description: Compute gradient by roberts operators
%
% Author: Haiying Liu
%   Date: Oct. 3, 2000
%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

dbstop if error

msg = nargchk(1, 1, nargin);
if (~isempty(msg))
  error(strcat('ERROR:', msg));
end

clear msg;

%======================================================================

% Construct roberts operator.
op_roberts = [1, 0; 0, -1] / sqrt(2);

% Get gradient.
bx = filter2(op_roberts, data, 'same');
by = filter2(rot90(op_roberts), data, 'same');

% Compute magnitude and direction.
theta        = 135 * pi / 180;
dx           = bx .* cos(theta) - by .* sin(theta);
dy           = bx .* sin(theta) + by .* cos(theta);
grad         = dx + i * dy;

gradient.mag = abs(grad);
% gradient.dir = angle(grad); % We don't need atan2
zeroDXMask = (dx==0);
nonZeroDXMask = ~zeroDXMask;
gradient.dir = zeros(size(data)); % this defines an image for directions
gradient.dir(nonZeroDXMask) = atan(dy(nonZeroDXMask)./dx(nonZeroDXMask));
gradient.dir(zeroDXMask) = pi/2; 

% Suppress non-maxima
gradient.mask = zeros(size(data));
nRow          = size(data, 1);
nCol          = size(data, 2);

for row = 2:nRow - 1
  for col = 2:nCol - 1
    wnd = grad(row - 1:row + 1, col - 1:col + 1);
    gradient.mask(row, col) = ...
      nonMaximaSuppress(real(wnd), imag(wnd));
  end
end
