% function [newClusterCenters, newClusterTrace, featureVectors] = seq13Background(rootName, nbFrames, showPictures)
%
% Usage: [newClusterCenters, newClusterTrace, featureVectors] = seq13Analysis(rootName, nbFrames, showPictures)
%
%
%if Output is cluster centers found from mean shift analysis 
% using starting points from the shot stack
% Tasks for space-time segmentation of shot stack
%
% 1. Read all frames and put in video stack
%
% 2. Make stack of grey images
%
% 3. Find 2 angle maps using Lucas-Kanade optical flow method
% For each pixel, find slope of edge and slope of projection direction.
% Find distance x of projection of this point on projection direction
%
% 4. Trim grey3D and rgb3D to match angleMaps
%
% 5.  Find two projection maps
%
% 6. Make two normalized color images and one illumination
%
% 7. Make feature vector list
%

% 5. Compute centroid in angle, projDist, R', G'  using the tagged pixels
%
% 6. Repeat 5 until fixed point
% 7. Make a feature list. Note here that we don't transpose before reshaping

% 7. Do 3-6 for every pixel

% 8. Regroup clusters that are close using centroids

% 9. For each pixel, keep track of which fixed point the regions starting at
% the pixel drifted to and which cluster they were absorbed into. Use
% the angle and gray level of this cluster to characterize the pixel.

% 10. Repeat process with the directions obtained at step 9 instead of the edge directions,
% until nothing changes

% 11. Use cluster center [ theta, projDist, R', G', Nb pixels per row] as feature vector for video shot, 
% since clusters characterize both locations of blobs, their colors and their motions.

% Also do a median filter on image, or use as pixel the pixel color for which the median angle was selected
% to prevent the background from being assigned a motion because of the median filtering of the angles
% which smears angles over the background

% =================================================================================

    	clear all;
    	rootName = '/fs/lamp/Projects/VACE/VACEData/LAMPSequences/Seq13/Input/Seq13_';
	initFrame = 360; endFrame = 375;
	top = 110; right = 295; bottom = 180; left = 213;

tic

% =================================================================================

% Put all parameters here, so it is clear what we assume
% magnitudeThreshold = 2;
% nbSegments = 10; % fixed number of segments, even when they are thinner than minNbPix
%angleRadius = 18;
%projRadius = 10;
%colorRadius = 25;

angleRange = 180;
minNbPix = 9; % segments with less than minNbPix are not used as features

% angleRadius = 6;
% projRadius = 8;
% rColorRadius = 6;
% gColorRadius = 6;
angleRadius = 20;
projRadius = 25;
rColorRadius = 10;
gColorRadius = 10;

% =================================================================================

% 1. Read all frames and put in video stack
disp('Read stack of color images');

	rgb3D = readSeq13Images(rootName, initFrame, endFrame, top, right, bottom, left);

% 2. Make stack of grey images
disp('Make stack of grey images');
grey3D = makeGreyImage3D(rgb3D);

% 3. Find angle maps
disp('Find angle maps');
[angleMapX, angleMapY, vDefined] = makeAngleMaps2(grey3D);
% [angleMapX, angleMapY, vDefined] = makeEdgeAngleMaps(grey3D);

[nbFrames, nbRows, nbCols] = size(angleMapX); % these are trimmed dimensions 
nbPix = nbFrames * nbRows * nbCols;
    

% 4. Trim grey3D and rgb3D to match angleMaps
disp('Trim stacks to size of angle maps');
[rgb3D, grey3D] = trimStacks(rgb3D, grey3D, nbFrames, nbRows, nbCols); 

if 1
% 4'. Quantize image to 32 color levels to speed up segmentation. *** CHECK
nbLevels = 32;
rgb3D = quantizeRGB(rgb3D, nbLevels);
end; % if

% 5. Find projection maps
disp('Find projection maps');
% projDistOffset  = ceil(sqrt(nbFrames^2 + nbRows^2 + nbCols^2)/2) + 1;
projDistOffsetX  = ceil(sqrt(nbFrames^2 + nbCols^2)/2) + 1;
projDistOffsetY  = ceil(sqrt(nbFrames^2 + nbRows^2)/2) + 1;

[projMapX, projMapY] = makeProjMaps2(angleMapX, angleMapY, projDistOffsetX, projDistOffsetY, vDefined);


% 6. Make two normalized color images and one illumination
disp('Make normalized color image stacks');
[rImage,gImage, illumImage] = getRGImages3D(rgb3D); % 3rd image is illumination part

% 7. Make feature vector list
disp('Make feature vector list');
featureList = zeros(nbPix, 6); % 8 images of same sizes

featureList(:,1) = reshape(angleMapX, nbPix,1); % transpose images so row order
featureList(:,2) = reshape(angleMapY, nbPix,1);
featureList(:,3) = reshape(projMapX, nbPix,1);
featureList(:,4) = reshape(projMapY, nbPix,1);
featureList(:,5) = reshape(rImage, nbPix,1);
featureList(:,6) = reshape(gImage, nbPix,1);
illum = reshape(illumImage, nbPix,1);
dataFlags = reshape(vDefined, nbPix,1);

% 8. Process only features where angles are defined
disp('Build nonzero feature list and trace');
[nonZeroFeatureList, nonZeroTrace] = keepNonZeroFeatures(featureList, dataFlags);

disp('Build nonzero illumination list');
nonZeroIllum = findNonZeroIllum(illum, nonZeroTrace);

% 9. Perform mean shift
disp('Make sparse feature matrices');
sparseFeatures = makeSparseFeatures(nonZeroFeatureList);
disp('Do mean shift');
tic
radii = [angleRadius, angleRadius, projRadius, projRadius, rColorRadius, gColorRadius];
[shiftList, newTrace] = findShiftedVectors(nonZeroFeatureList, sparseFeatures, nonZeroTrace, radii);
toc
disp('Find weights of mean shift centers');
weights = findWeights(size(shiftList,1), newTrace); 

% 10. Group neighbors; *** CHECK full radii instead
disp('Group neighbors');
% [clusterCenters, clusterTrace, clusterWeights] = groupNeighbors(shiftList, newTrace, weights, round(radii/2));
[clusterCenters, clusterTrace, clusterWeights] = groupNeighbors(shiftList, newTrace, weights, round(radii));

disp('Find cluster illuminations by averaging illums of pixels contributing to neigbor feature vectors');
clusterIllum = findClusterIllum(nonZeroIllum, clusterCenters, sparseFeatures, radii);

% 11. Find segments
% Make clusterTrace a 3D stack giving cluster label for each pixel of stack
disp('Label pixels by reshaping cluster trace to video stack shape');
clusterTrace3D = reshape(clusterTrace, nbFrames, nbRows, nbCols);

% Find line segments corresponding to clusters
% Use cluster trace to find only portions of cluster lines for which there are contributing pixels in the same row
disp('Find line segments along cluster lines');
clusterSegments =  findClusterSegments3D(clusterCenters, clusterTrace3D);

% Segment shot stack by giving pixels the values of the clusters which they clustered to
disp('Find color and angle segmentations');
[segmented3D, angle3DX, angle3DY] = segmentShotsFromClusters3D(clusterCenters, clusterTrace3D, clusterIllum);

% Keep only clusters for which average nb of pixels per row is large
disp('Sort clusters by stripe thickness');
[bestClusterCenters, bestClusterSegments, clusterTraceBest] = pickBestClusters3D(clusterCenters, clusterSegments, minNbPix);
bestClusterIllum = findBestClusterIllum(clusterIllum, clusterTraceBest);
bestClusterTrace = mergeTraces(clusterTrace, clusterTraceBest);
bestClusterTrace3D = reshape(bestClusterTrace, nbFrames, nbRows, nbCols);
[bestSegmented3D, bestAngle3DX, bestAngle3DY] = segmentShotsFromClusters3D(bestClusterCenters, bestClusterTrace3D, bestClusterIllum);

% Add average nbs of pixels per frame that contributed to them
% and remove offset from distX and distY
disp('Use stripe thickness instead of cluster mass as feature');
featureVectors = makeFeatureVectors3D(bestClusterCenters, bestClusterSegments, projDistOffsetX, projDistOffsetY)

disp('Save wokspace as a *.mat file');
save spaceTime3D_Seq13;

toc

% Now use plot routines to show the lines corresponding to moving objects
% For now we use 2D projections of lines on x-t and y-t images
% Find way to draw these lines with the right colors in perspective projections
% using intersection between lines and cube boundaries

if 1
showPictures = 1;
	if showPictures
	  showThickness = 0;
	  [allLinesImX, allLinesImY, allLinesTop]  = drawClusterLines3D(nbFrames, nbRows, nbCols,clusterCenters, clusterSegments, clusterIllum, showThickness);
	  [bestLinesImX1, bestLinesImY1, bestLinesTop]  = drawClusterLines3D(nbFrames, nbRows, nbCols,bestClusterCenters, bestClusterSegments, bestClusterIllum, showThickness);
	  showThickness = 1;
	  [bestLinesImX2, bestLinesImY2, bestLinesTop]  = drawClusterLines3D(nbFrames, nbRows, nbCols,bestClusterCenters, bestClusterSegments, bestClusterIllum, showThickness);
	
	  showAllImages3D(rgb3D, allLinesImX, allLinesImY, ...
	  		bestLinesImX1, bestLinesImY1, bestLinesImX2, bestLinesImY2, bestLinesTop, ...
			angleMapX, angleMapY, projMapX, projMapY, segmented3D, angle3DX, angle3DY);
	
	end
end
