/*=================================================================
 *
 * fastScaledCentroids.c  
 *
 * Usage: centroids = fastScaledCentroids(counts, neighborIndices, distances, featureVects, weights, oldCentroids, invVariances);
 * Inputs:
 * 1. Counts of feature vectors within range of points being shifted
 * 2. Indices of background feature vectors within range
 * 3. Distances of background feature vectors within range
 * 4. Feature vectors of background
 * 5. Weights for these feature vectors
 * 6. Old positions of centroids
 * 7. Inverses of variances along each feature and for each centroid
 *
 * Outputs:
 * Positions of centroids
 *
 * This is compiled to a MEX-file for MATLAB.  
 * Copyright 2001-2002 Daniel DeMenthon & University of Maryland
 * Gaussian weighting added 04-2002 to original fastCentroids.c 
 * Remi Megret : Variance based gaussian weighting added 05-2002
 *
 *=================================================================*/


#include <math.h>
#include "mex.h"

/********************************************************************************/
/*******************************************************************************/

void mexFunction( int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[] )

{
  int i, j, k;
  int nbCentroids, nbFeatureVects, nbFeatures, nbPtsInSphere, neighborIndex;
  int vectCoordIndex, centroidCoordIndex;

  double totalWeight, neighborWeight, neighborDistance, vectCoord, sigma, twoSigmaSquared;
  double  *weights, *counts, *featureVects, *centroids, *neighbors, *distances, *oldCentroids;

  double *weightTimesCentroid, *centroidCoordPtr, *oldCentroidCoordPtr;
  
  mxArray *neighborCell, *distanceCell;

  double delta;
  double *invVariances;
  double gaussianWeight;

  if(nrhs != 7){
    mexErrMsgTxt("Seven inputs required: counts, neighborIndices, distances, featureVects, weights, oldCentroids, invVariances");
  }
  if(nlhs != 1){
    mexErrMsgTxt("One output required: centroids");
  }

  nbCentroids = mxGetM(prhs[0]); /* nb of moving centroids */
  counts = mxGetPr(prhs[0]);
  
  nbFeatureVects = mxGetM(prhs[3]); /* nb of background points */
  nbFeatures = mxGetN(prhs[3]); /* nb of feature components */
  featureVects = mxGetPr(prhs[3]);
  
  // neighborIndices = prhs[1]; /* we need to work with mxArray to get to cells */
  weights = mxGetPr(prhs[4]);
  oldCentroids = mxGetPr(prhs[5]);
  invVariances = mxGetPr(prhs[6]);
  //sigma = mxGetScalar(prhs[6]);

  if (mxGetM(prhs[6]) != nbCentroids) {
    mexErrMsgTxt("invVariances should have as many lines as centroids");
  }
  if (mxGetN(prhs[6]) != nbFeatures) {
    mexErrMsgTxt("invVariances should have as many columns as there are features");
  }

  weightTimesCentroid = (double *) mxCalloc(nbFeatures, sizeof(double)); 
  plhs[0] = mxCreateDoubleMatrix(nbCentroids, nbFeatures, mxREAL);
  centroids = mxGetPr(plhs[0]);
  
  for(i=0;i<nbCentroids;i++){
	// mexPrintf("i %d\n", i);
    nbPtsInSphere = counts[i]; // there is sometimes no points in there
    neighborCell = mxGetCell(prhs[1], i); // mxArray for cell of neighborIndices
    neighbors = mxGetPr(neighborCell);

    totalWeight = 0; // init
    for(j=0;j<nbFeatures;j++){ // init coords to zero
      weightTimesCentroid[j] = 0.0;
    }
    for(k=0;k<nbPtsInSphere;k++){
	  //   mexPrintf("k %d\n", k);
      neighborIndex = neighbors[k] -1; // - 1 as arrays here are zero-based

      /* Gaussian weighting, with only diagonal variances */
      gaussianWeight = 0;
      for(j=0;j<nbFeatures;j++){ // init coords to zero
	centroidCoordIndex = j * nbCentroids + i;
	vectCoordIndex = j * nbFeatureVects + neighborIndex;
	delta = featureVects[vectCoordIndex] - oldCentroids[centroidCoordIndex];
	// invVariances should have the same size as oldCentroids
	gaussianWeight += delta * delta * invVariances[centroidCoordIndex];
      }
      gaussianWeight = exp(-gaussianWeight/2.0);

      /* Weight for neighbor k */
      neighborWeight = weights[neighborIndex] * gaussianWeight;
      // neighborWeight = weights[neighborIndex] * exp(-(neighborDistance * neighborDistance)/twoSigmaSquared);

      totalWeight += neighborWeight;
      for(j=0;j<nbFeatures;j++){      
	//	mexPrintf("j %d\n", j);
	vectCoordIndex = j * nbFeatureVects + neighborIndex;
	//	mexPrintf("vectCoordIndex %d\n", vectCoordIndex);
	vectCoord = *(featureVects + vectCoordIndex);
	//	mexPrintf("vectCoord %f\n", vectCoord);
	weightTimesCentroid[j] += neighborWeight * vectCoord;
      }
    }
    for(j=0;j<nbFeatures;j++){
      centroidCoordIndex = j * nbCentroids + i;
      centroidCoordPtr = centroids + centroidCoordIndex;
      if(totalWeight == 0){
		oldCentroidCoordPtr = oldCentroids + centroidCoordIndex;
		*centroidCoordPtr = *oldCentroidCoordPtr;
      }
      else{
		//	mexPrintf("% totalWeight %f\n", totalWeight);
		*centroidCoordPtr = weightTimesCentroid[j]/totalWeight;
      }
    }
  }

  mxFree(weightTimesCentroid);
}

/*******************************************************************************/
