function [clusterCenters, clusterTrace, closeVectIndexList, ranges] = meanShiftMultiPass3(featureList, dataFlags, radius, importances, doGaussian)
%
% Usage: [clusterCenters, clusterTracecloseVectIndexList, ranges] = meanShiftMultiPass3(featureList, dataFlags, radius, importances)
%
% This is a general mean shift calculation with no assumption on the meaning of the features
% We are using fastMergeDups, a merging of duplicates based on sorting 
% which is faster than the one using range_search for data with a lot of preexisting clusters
%
% clusterTrace is the list of indices showing which cluster each feature list item belongs to
% Therefore length(clusterTrace) = length(featureList)
% and max(clusterTrace) = length(clusterCenters)
%
% clusterCenters is a matrix N x n, with N number of clusters and rows showing their components
%
% dataFlags show which feature vectors can be used (here some don't have angle components
% In this version we group neighbors only once
%
% Functions used by meanShiftEngine3:
%
%   nn_prepare, range_search from TSTools
%   keepNonZeroFeatures
%   normalizeFeatures3
%   fastMergeDups
%   fastCompleteMergeDups
%   mergeTraces
%   findTreeShiftedVectors4
%   shiftTreeFeatures
%   findWeights
%   findTreeNeighbors 
%   inverseMapFeatures2
%

% deltaRadius = radius/nbSteps; % we increment radius by deltaRadius at every step
% We multiply initial radius by 2 at every iteration step

disp('Build nonzero feature list and trace');
[nonZeroFeatureList, nonZeroTrace] = keepNonZeroFeatures(featureList, dataFlags);

disp('Map ranges of each component to range [0,1] (map to hypercube of size 1)');
[normalizedFeatures, ranges] = normalizeFeatures(double(nonZeroFeatureList), importances);

disp('Find weighted background of features');
tic
quantizError = 1/100;
[weightedFeatures, featureWeights, tempTrace] = fastMergeDups(normalizedFeatures, quantizError);
toc  
weightedTrace = mergeTraces(nonZeroTrace, tempTrace);

iteratedRadius = radius/10;
nbSteps = 20;
% factor = 2;
% factor = 1.5;
% factor = 1.35;
factor = 1.25;

clusters = weightedFeatures;
weights = featureWeights;
trace = weightedTrace;

for i = 1:nbSteps
    iteratedRadius = factor * iteratedRadius;
     if iteratedRadius > radius
        iteratedRadius = radius;
    end;   
    disp(['Iterated radius at step ',num2str(i),' is ', num2str(iteratedRadius)]);
    quantizError = iteratedRadius/20;
    treeData = nn_prepare(clusters);
    if i == 1
        firstTreeData = treeData; % treeData for weightedFeatures
    end;
	[newClusters, newTrace] = findTreeShiftedVectors(clusters, weights, treeData, iteratedRadius, quantizError, doGaussian);
    trace = mergeTraces(trace, newTrace);
    newWeights = findWeights(trace);

    if iteratedRadius == radius % exit
        break;
    else
 	    clusters = newClusters; % so that when we exit, clusters and newClusters remain different
        weights = newWeights;
    end;
end;


% Methods give mostly same results because there are very few neighbors to choose from when getting max
disp(' Group neighbors ');
if 1
   [newClusters, localTrace] = findMaxNeighbors(newClusters, newWeights, radius); 

else
    disp('Find cluster densities using nearest neighbors');
% 	clusterDensities0 = findClusterDensities(newClusters, clusters, weights, treeData, radius);
    clusterDensities = findClusterDensities(newClusters, weightedFeatures, featureWeights, firstTreeData, radius);
   [newClusters, localTrace] = findMaxNeighbors(newClusters, clusterDensities, radius/2); % this favors clusters that have the most contributions   
end;
newTrace = mergeTraces(trace, localTrace);


disp('Find vector of featureList that is closest to each clusterCenter');
% Use color of nearest neighbor featureList vector as color of cluster
% This should be a cleaner color than average color

closeVectIndexList = findCloseVects(weightedFeatures, weightedTrace, newClusters, firstTreeData);

clusterTrace = newTrace;
disp('Denormalize clusters');
clusterCenters = inverseMapFeatures(newClusters, ranges, importances);




