function [clusterCenters, clusterTrace, closeVectIndexList, ranges, allTraces, regionCounts] = ...
			meanShiftMultiPassTraces(featureList, dataFlags, importances, doGaussian)
%
% Usage: [clusterCenters, clusterTracecloseVectIndexList, ranges, allTraces, regionCounts] = meanShiftMultiPassTraces(featureList, dataFlags, radius, importances)
%
% Data-collecting version of meanShiftMultiPass3
% We collect traces at each clustering step to build a tree of coarse-to-fine segmentation
% Also, we collect number of regions at each step to be able to trace a curve nbRegions = f(radius)

% This is a general mean shift calculation with no assumption on the meaning of the features
% We are using fastMergeDups, a merging of duplicates based on sorting 
% which is faster than the one using range_search for data with a lot of preexisting clusters
%
% clusterTrace is the list of indices showing which cluster each feature list item belongs to
% Therefore length(clusterTrace) = length(featureList)
% and max(clusterTrace) = length(clusterCenters)
%
% clusterCenters is a matrix N x n, with N number of clusters and rows showing their components
%
% dataFlags show which feature vectors can be used (here some don't have angle components
% In this version we group neighbors only once
%
% Functions used by meanShiftEngine3:
%
%   nn_prepare, range_search from TSTools
%	keepNonZeroFeatures
%   normalizeFeatures
%   fastMergeDups
%   mergeTraces
%	findTreeShiftedVectors2
%   shiftTreeFeatures
%   mergeShiftFlags
%	findWeights
%   groupTreeNeighbors
%   findTreeNeighbors 
%   inverseMapFeatures
%

% deltaRadius = radius/nbSteps; % we increment radius by deltaRadius at every step
% We multiply initial radius by 2 at every iteration step

% In this data-collecting version, keep traces

disp('Map ranges of each component to range [0,1] (map to hypercube of size 1)');
% reverse order to build initial tree on initial pixels
[normalizedFeatures, ranges] = normalizeFeatures3(double(featureList), importances);
initTreeData = nn_prepare(normalizedFeatures);

disp('Build nonzero feature list and trace');
[nonZeroFeatureList, nonZeroTrace] = keepNonZeroFeatures(normalizedFeatures, dataFlags);

disp('Find weighted background of features');
tic
quantizError = 1/100;
[weightedFeatures, featureWeights, tempTrace] = fastMergeDups(nonZeroFeatureList, quantizError);
toc  
weightedTrace = mergeTraces(nonZeroTrace, tempTrace);

allTraces = weightedTrace; % first mapping is after regrouping duplicates
radius = 0.5;
nbSteps = 20;
% factor = 2;
% factor = 1.5;
factor = 1.25;
% factor = 1.1;
minNbClusters = 15; 

initRadius = radius/(factor^(nbSteps-1)); % to get radius at end of loop
clusters = weightedFeatures;
weights = featureWeights;
trace = weightedTrace;
iteratedRadius = initRadius;

regionCounts = [0, max(nonZeroTrace)]; % initial radius and initial nb of regions
for i = 1:nbSteps
    disp(['Iterated radius at step ', num2str(i),' is ', num2str(iteratedRadius)]);
    quantizError = iteratedRadius/20;
    treeData = nn_prepare(clusters);

    [newClusters, newTrace] = findTreeShiftedVectors4(clusters, weights, treeData, iteratedRadius, quantizError, doGaussian);
    reductionFactor = length(trace)/length(newTrace)
    trace = mergeTraces(trace, newTrace);
    allTraces = [allTraces, trace]; % recording data before merging
    newWeights = findWeights(trace);
   
    if max(trace)< minNbClusters
        radius = iteratedRadius; % so that we use right radius in findMaxNeightbors
        break;
    end

    regionCounts = [regionCounts; [iteratedRadius, max(trace)]]; % recording data

    clusters = newClusters; % so that when we exit, clusters and newClusters remain different
    weights = newWeights;
    iteratedRadius = factor * iteratedRadius;
end;

disp(' Group neighbors ');
% Methods give mostly same results because there are very few neighbors to choose from when getting max
if 1
    [newClusters, localTrace] = findMaxNeighbors(newClusters, newWeights, radius/2); % this favors clusters that have the most contributions
%     [newClusters, localTrace] = findMaxNeighbors(newClusters, newWeights, radius); % this favors clusters that have the most contributions
else
    disp('Find cluster densities using nearest neighbors');
% 	clusterDensities0 = findClusterDensities(newClusters, clusters, weights, treeData, radius);
    clusterDensities = findClusterDensities(newClusters, weightedFeatures, featureWeights, initTreeData, radius);
   [newClusters, localTrace] = findMaxNeighbors(newClusters, clusterDensities, radius/2); % this favors clusters that have the most contributions   
end;
newTrace = mergeTraces(trace, localTrace);


disp('Find vector of featureList that is closest to each clusterCenter');
% Use color of nearest neighbor featureList vector as color of cluster
% This should be a cleaner color than average color

closeVectIndexList = findCloseVects2(normalizedFeatures, newClusters, initTreeData);

clusterTrace = newTrace;
disp('Denormalize clusters');
clusterCenters = inverseMapFeatures2(newClusters, ranges, importances);




