function [clusterCenters, clusterTrace, closeVectIndexList, ranges] = ...
			meanShifOptimalTrace(featureList, dataFlags, importances, doGaussian)
%
% Usage: [clusterCenters, clusterTrace, closeVectIndexList] = ...
%         meanShifOptimalTrace(featureList, dataFlags, importances, doGaussian)
%
% Combination of meanShiftMultiPass and findOptimalClustering
% % Return a trace in which each region corresponds to the longest living clusters when radius is increased
% and the corresponding clusters

% Check if there was a grouping between parent trace and child trace
% if there is, relabel new cluster with a new global label

% We collect traces at each clustering step to build a tree of coarse-to-fine segmentation
% Also, we collect number of regions at each step to be able to trace a curve nbRegions = f(radius)

% This is a general mean shift calculation with no assumption on the meaning of the features
% We are using fastMergeDups, a merging of duplicates based on sorting 
% which is faster than the one using range_search for data with a lot of preexisting clusters
%
% clusterTrace is the list of indices showing which cluster each feature list item belongs to
% Therefore length(clusterTrace) = length(featureList)
% and max(clusterTrace) = length(clusterCenters)
%
% clusterCenters is a matrix N x n, with N number of clusters and rows showing their components
%
% dataFlags show which feature vectors can be used (here some don't have angle components
% In this version we group neighbors only once
%
% Functions used by meanShiftEngine3:
%
%   nn_prepare, range_search from TSTools
%	keepNonZeroFeatures
%   normalizeFeatures
%   fastMergeDups
%   mergeTraces
%	findTreeShiftedVectors2
%   shiftTreeFeatures
%   mergeShiftFlags
%	findWeights
%   groupTreeNeighbors
%   findTreeNeighbors 
%   inverseMapFeatures
%

% deltaRadius = radius/nbSteps; % we increment radius by deltaRadius at every step
% We multiply initial radius by 2 at every iteration step

% In this data-collecting version, keep traces
disp('Build nonzero feature list and trace');
[nonZeroFeatureList, nonZeroTrace] = keepNonZeroFeatures(featureList, dataFlags);

disp('Map ranges of each component to range [0,1] (map to hypercube of size 1)');
% It is better to normalize after removing the zeros, for which data are unknown
[normalizedFeatures, ranges] = normalizeFeatures(double(nonZeroFeatureList), importances);

disp('Find weighted background of features');

quantizError = 1/100;
[weightedFeatures, featureWeights, tempTrace] = fastMergeDups(normalizedFeatures, quantizError);
  
weightedTrace = mergeTraces(nonZeroTrace, tempTrace);

% radius = 0.5;
radius = 0.78;
% nbSteps = 20;
nbSteps = 18;
% nbSteps = 16;
% factor = 2;
% factor = 1.5;
factor = 1.25;
% factor = 1.1;

initRadius = radius/(factor^(nbSteps-1)); % to get radius at end of loop
clusters = weightedFeatures;
weights = featureWeights;
trace = weightedTrace;
iteratedRadius = initRadius;
initNbClusters = 1000; % max nb of clusters for segmentation
savedFirstTrace = 0;

for i = 1:nbSteps
    disp(['Iterated radius at step ', num2str(i),' is ', num2str(iteratedRadius)]);
    quantizError = iteratedRadius/20;
    treeData = nn_prepare(clusters);

    [newClusters, newTrace] = findTreeShiftedVectors(clusters, weights, treeData, iteratedRadius, quantizError, doGaussian);
    trace = mergeTraces(trace, newTrace);
    newWeights = findWeights(trace);
    
    if i == 1 % first pass through loop
        firstTreeData = treeData; % treeData for finding nearest neigbors at end of code
    end
    
    nbClusters = max(trace)
    if (nbClusters<=initNbClusters)
		if 0
			disp(' Group neighbors '); % try without this step too to speed things up
			[newClusters, localTrace] = findMaxNeighbors(newClusters, newWeights, iteratedRadius/2); 
			trace = mergeTraces(trace, localTrace);
		end;
   	    parentTrace = trace;
        if ~savedFirstTrace
			bestTrace = trace; % this sets zeros for pixels with no clustering
            bestClusters = newClusters;
			nbBestClusters = max(bestTrace);
            disp(['>>> nb of initial clusters = ',num2str(nbBestClusters),' <<<<']);
			bestAges = ones(nbBestClusters,1); % ages of best clusters
			ages = ones(nbBestClusters,1); % ages of initial clusters
		    allocatedLabels = nbBestClusters; % every cluster has distinct label in first trace
            isBestTraceUpdated = zeros(nbBestClusters,1);
            savedFirstTrace = 1;
        else % after first pass through previous if
			nbParentClusters = max(parentTrace); % note: this would not work if we relabeled all traces
			for j = 1:nbParentClusters % note: max(parentTrace) changes within this loop
%                 j
                localCluster = newClusters(j,:);
				parentIndices = find(parentTrace == j); % pixels that have label j in parents
				childRegionLabels = childTrace(parentIndices); % find cluster labels in child for those pixels
                maxLabel = max(childRegionLabels);
                minLabel = min(childRegionLabels);
				if maxLabel == minLabel % then single label in child, no grouping, age of cluster is increased
					clusterLabel = maxLabel;
					parentTrace(parentIndices) = clusterLabel; % replace parent label with index of child, so we can track age
                    ages(clusterLabel) = ages(clusterLabel) + 1;
					if isBestTraceUpdated(clusterLabel)
						bestAges(clusterLabel) = bestAges(clusterLabel) + 1;            
                    else % bestTrace has not be updated yet for this cluster
                        bestLocalLabels = bestTrace(parentIndices);
                        bestLocalAges = bestAges(bestLocalLabels);
                        oldestAge = max(bestLocalAges); % find oldest cluster overlapped by maxLabel cluster
			            if ages(clusterLabel) > oldestAge
% 	                	if ages(clusterLabel) > sqrt(oldestAge)
%  	                	if ages(clusterLabel) > log(oldestAge) - 1
                           bestClusters(clusterLabel,:) = localCluster;
                    		bestTrace(parentIndices) = clusterLabel; % place clusterLabel in pixels of bestTrace
                            bestAges(clusterLabel) = ages(clusterLabel);
                    		isBestTraceUpdated(clusterLabel) = 1;
                    	end;
                    end;
                else  % a grouping occurred
                    allocatedLabels = allocatedLabels + 1;
                    parentTrace(parentIndices) = allocatedLabels; % a new label for the grouping
                    ages = [ages;1]; % no  point checking ages of bestTraces
                    isBestTraceUpdated = [isBestTraceUpdated; 0]; 
                end; % if maxLabel == minLabel
			end; % for j
        end % else isSavedTrace
        childTrace = parentTrace;
    end % if < initNbClusters

    clusters = newClusters; % so that when we exit, clusters and newClusters remain different
    weights = newWeights;
    iteratedRadius = factor * iteratedRadius;
end; % for i

disp(['@@@@@@@ Oldest region age is ',num2str(max(bestAges)), ' @@@@@@@']);

disp('Reassign bestTrace labels');
% Labels in bestTrace have missing elements in label sequence;
% replace current labels using a continuous sequence from 1 to nb of labels

bestTrace2 = bestTrace;
% bestClusters2 = bestClusters;

maxLabel = max(bestTrace);
labelMapping = zeros(maxLabel);
availableLabel = 0;
bestClusters2 = [];

for i = 1:length(bestTrace)
    label = bestTrace(i);
    if label ~= 0
        newLabel = labelMapping(label);
        if newLabel == 0 % was not reassigned yet
            availableLabel = availableLabel + 1;
            bestClusters2 = [bestClusters2; bestClusters(label,:)];
            labelMapping(label) = availableLabel;
            bestTrace2(i) = availableLabel;
        else
            bestTrace2(i) = newLabel;
        end
    end
end

% disp(' Group neighbors '); % there is no single radius for the clusters, so we cannot connect neighborh clusters anymore
% [newClusters, localTrace] = findMaxNeighbors(newClusters, newWeights, radius); % this favors clusters that have the most contributions
% newTrace = mergeTraces(trace, localTrace);


disp('Find vector of featureList that is closest to each clusterCenter');
% Use color of nearest neighbor featureList vector as color of cluster
% This should be a cleaner color than average color

closeVectIndexList = findCloseVects(weightedFeatures, weightedTrace, bestClusters2, firstTreeData);

clusterTrace = bestTrace2;
disp('Denormalize clusters');
clusterCenters = inverseMapFeatures(bestClusters2, ranges, importances);




