/*=================================================================
 *
 * fastCentroids.c  
 *
 * Usage: centroids = fastCentroids(counts, neighborIndices, featureVects, weights, oldCentroids);
 * Inputs:
 * 1. Counts of feature vectors within range of points being shifted
 * 2. Indices of background feature vectors within range
 * 3. Feature vectors of background
 * 4. Weights for these feature vectors
 *
 * Outputs:
 * Positions of centroids
 *
 * This is a MEX-file for MATLAB.  
 * Copyright 2001 Daniel DeMenthon & University of Maryland 
 *
 *=================================================================*/


#include <math.h>
#include "mex.h"

/********************************************************************************/
/*******************************************************************************/

void mexFunction( int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[] )

{
  int i, j, k;
  int nbCentroids, nbFeatureVects, nbFeatures, nbPtsInSphere, neighborIndex;
  int vectCoordIndex, centroidCoordIndex;

  double totalWeight, neighborWeight, vectCoord;
  double  *weights, *counts, *featureVects, *centroids, *neighbors, *oldCentroids;

  double *weightTimesCentroid, *centroidCoordPtr, *oldCentroidCoordPtr;
  
  mxArray *neighborCell, *neighborIndices;

  if(nrhs != 5){
    mexErrMsgTxt("Four inputs required: counts, neighborIndices, featureVects, weights, oldCentroids");
  }
  if(nlhs != 1){
    mexErrMsgTxt("One output required: centroids");
  }

  nbCentroids = mxGetM(prhs[0]); /* nb of moving centroids */
  nbFeatureVects = mxGetM(prhs[2]); /* nb of background points */
  nbFeatures = mxGetN(prhs[2]); /* nb of feature components */
  weightTimesCentroid = (double *) mxCalloc(nbFeatures, sizeof(double)); 
  
  counts = mxGetPr(prhs[0]);
  // neighborIndices = prhs[1]; /* we need to work with mxArray to get to cells */
  featureVects = mxGetPr(prhs[2]);
  weights = mxGetPr(prhs[3]);
  oldCentroids = mxGetPr(prhs[4]);

  plhs[0] = mxCreateDoubleMatrix(nbCentroids, nbFeatures, mxREAL);
  centroids = mxGetPr(plhs[0]);


  for(i=0;i<nbCentroids;i++){
 // mexPrintf("i %d\n", i);
    nbPtsInSphere = counts[i]; // there is sometimes no points in there
    neighborCell = mxGetCell(prhs[1], i); // mxArray for cell of neighborIndices
    neighbors = mxGetPr(neighborCell);
    totalWeight = 0; // init
    for(j=0;j<nbFeatures;j++){ // init coords to zero
      weightTimesCentroid[j] = 0.0;
    }
    for(k=0;k<nbPtsInSphere;k++){
 //   mexPrintf("k %d\n", k);
      neighborIndex = neighbors[k] -1; // - 1 as arrays here are zero-based
      neighborWeight = weights[neighborIndex];
      totalWeight += neighborWeight;
      for(j=0;j<nbFeatures;j++){      
 //	mexPrintf("j %d\n", j);
	vectCoordIndex = j * nbFeatureVects + neighborIndex;
 //	mexPrintf("vectCoordIndex %d\n", vectCoordIndex);
	vectCoord = *(featureVects + vectCoordIndex);
 //	mexPrintf("vectCoord %f\n", vectCoord);
	weightTimesCentroid[j] += neighborWeight * vectCoord;
      }
    }
    for(j=0;j<nbFeatures;j++){
      centroidCoordIndex = j * nbCentroids + i;
      centroidCoordPtr = centroids + centroidCoordIndex;
      if(totalWeight == 0){
	oldCentroidCoordPtr = oldCentroids + centroidCoordIndex;
	*centroidCoordPtr = *oldCentroidCoordPtr;
      }
      else{
 //	mexPrintf("% totalWeight %f\n", totalWeight);
	*centroidCoordPtr = weightTimesCentroid[j]/totalWeight;
      }
    }
  }

  mxFree(weightTimesCentroid);
}

/*******************************************************************************/
