/*=================================================================
 *
 * fastClusterMax.c  
 *
 * Usage: [clusterCenters, clusterTrace, nbClusters] = fastClusterMax(shiftList, clusterDensities, neighborIndices);
 * Inputs:
 * 1. shiftList, List of centroids found by mean shift
 * 2. clusterDensities, reflecting numbers of feature points in neighborhoods of clusters
 * 3. Cell array of neighbors of each pt in shiftList
  *
 * Outputs:
 * 1. clusterCenters: List of clusters grouping neighbors in shiftList
 * 2. clusterTrace: Index mapping from shiftList to clusterCenters
 * 3. nbClusters: Nb of clusters, used to prune arrays back in Matlab
 *
 * Among neighbor clusters, we select one true peak, one with highest density around it
 * Here we do not find centroid between neighbor peaks of same height
 * But just pick one, as they are typically only 2 or 3 candidates
 *
 * This is a MEX-file for MATLAB.  
 * Copyright 2001 Daniel DeMenthon & University of Maryland 
 *
 *=================================================================*/

#include <math.h>
#include "mex.h"

//=========================================================================

void mexFunction( int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[] )     
{
  int i, j, k;
  int nbPts, nbFeatures, clusterIndex, clusterCount, stackCount;
  int indexListCount, nbNeighbors, localIndex, neighborIndex;
  int clusterCoordIndex, peakCoordIndex, peakIndex;
  double density, maxDensity;
  int *usedPtFlags, *indexList, *stack;
  double *shiftList, *clusterCenters, *clusterDensities;
  double *clusterTrace, *nbClusters, *localNeighborIndices;
  double  *clusterCoordPtr;
  mxArray *neighborCell;
  
  nbPts = mxGetM(prhs[0]); /* nb of pts in shiftList */
  nbFeatures = mxGetN(prhs[0]); /* nb of feature components */

  shiftList = mxGetPr(prhs[0]);
  clusterDensities = mxGetPr(prhs[1]);
  
  plhs[0] = mxCreateDoubleMatrix(nbPts, nbFeatures, mxREAL);
  clusterCenters = mxGetPr(plhs[0]);

  plhs[1] = mxCreateDoubleMatrix(nbPts, 1, mxREAL);
  clusterTrace = mxGetPr(plhs[1]);

  plhs[2] = mxCreateDoubleMatrix(1, 1, mxREAL);
  nbClusters = mxGetPr(plhs[2]);

  usedPtFlags = (int *) mxCalloc(nbPts, sizeof(int)); /* initialize to zeros */
  indexList = (int *) mxCalloc(nbPts, sizeof(int)); /* initialize to zeros */
  stack = (int *) mxCalloc(nbPts, sizeof(int)); /* initialize to zeros */
  
  clusterCount = 0;
  for(i=0;i<nbPts;i++){
    if(!usedPtFlags[i]){
      usedPtFlags[i] = 1;
      for(j=0;j<nbPts;j++){
      	indexList[j] = 0;
      	stack[j] = 0;
      }
	  indexListCount = 0;
      stackCount = 0;
      clusterIndex = i;
      indexList[indexListCount++] = clusterIndex + 1; // include unused pt in list
      
      while(1){// explore neighborhoods of neighbors of clusterIndex
		neighborCell = mxGetCell(prhs[2], clusterIndex); // find neighbors
		nbNeighbors = mxGetN(neighborCell); // indices are stored on single row, M=1
		localNeighborIndices = mxGetPr(neighborCell);
		for(k=0;k<nbNeighbors;k++){
		  localIndex = localNeighborIndices[k] - 1;// -1 as arrays here are zero-based
		  if(!usedPtFlags[localIndex]){
		  	usedPtFlags[localIndex] = 1;
		  	stack[stackCount++] = localIndex; // stack
		  	indexList[indexListCount++] = localIndex + 1;
		  }
		}// end for k
		if(stackCount>0){
			stackCount--; // we had incremented stackCount after storing
			clusterIndex = stack[stackCount]; // examine new neighbor
		}
		else{
			break; // get out of while
	    }
	  }// end while
	  // Now find centroid of these neighbors
	  maxDensity = 0;

	  for(k=0;k<indexListCount;k++){
	  	neighborIndex = indexList[k] - 1;// arrays here are zero-based
	  	clusterTrace[neighborIndex] = clusterCount + 1;
		density = clusterDensities[neighborIndex];
		if(density > maxDensity){
		  maxDensity = density;
		  peakIndex = neighborIndex; // we pick up index for a cluster of max density
		}
	  }
	  for(j=0;j<nbFeatures;j++){
	    clusterCoordIndex = j * nbPts + clusterCount;
	    peakCoordIndex = j * nbPts + peakIndex;
	    clusterCoordPtr = clusterCenters + clusterCoordIndex;
	    *clusterCoordPtr = *(shiftList + peakCoordIndex);
	  }
	  clusterCount++;
    }// end if usedPtFlags
  }// end for i
  
  *nbClusters = clusterCount;
  
  mxFree(usedPtFlags);
  mxFree(indexList);
  mxFree(stack);
}
