/*=================================================================
 *
 * fastGaussCentroids.c  
 *
 * Usage: centroids = fastGaussCentroids(counts, neighborIndices, distances, featureVects, weights, oldCentroids, sigma);
 * Inputs:
 * 1. Counts of feature vectors within range of points being shifted
 * 2. Indices of background feature vectors within range
 * 3. Distances of background feature vectors within range
 * 4. Feature vectors of background
 * 5. Weights for these feature vectors
 * 6. Standard deviation of gaussian weighting
 *
 * Outputs:
 * Positions of centroids
 *
 * This is compilled to a MEX-file for MATLAB.  
 * Copyright 2001-2002 Daniel DeMenthon & University of Maryland
 * Gaussian weighting added 04-2002 to original fastCentroids.c 
 *
 *=================================================================*/


#include <math.h>
#include "mex.h"

/********************************************************************************/
/*******************************************************************************/

void mexFunction( int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[] )

{
  int i, j, k;
  int nbCentroids, nbFeatureVects, nbFeatures, nbPtsInSphere, neighborIndex;
  int vectCoordIndex, centroidCoordIndex;

  double totalWeight, neighborWeight, neighborDistance, vectCoord, sigma, twoSigmaSquared;
  double  *weights, *counts, *featureVects, *centroids, *neighbors, *distances, *oldCentroids;

  double *weightTimesCentroid, *centroidCoordPtr, *oldCentroidCoordPtr;
  
  mxArray *neighborCell, *distanceCell;

  if(nrhs != 7){
    mexErrMsgTxt("Seven inputs required: counts, neighborIndices, distances, featureVects, weights, oldCentroids, sigma");
  }
  if(nlhs != 1){
    mexErrMsgTxt("One output required: centroids");
  }

  nbCentroids = mxGetM(prhs[0]); /* nb of moving centroids */
  counts = mxGetPr(prhs[0]);
  
  nbFeatureVects = mxGetM(prhs[3]); /* nb of background points */
  nbFeatures = mxGetN(prhs[3]); /* nb of feature components */
  featureVects = mxGetPr(prhs[3]);
  
  // neighborIndices = prhs[1]; /* we need to work with mxArray to get to cells */
  weights = mxGetPr(prhs[4]);
  oldCentroids = mxGetPr(prhs[5]);
  sigma = mxGetScalar(prhs[6]);

  weightTimesCentroid = (double *) mxCalloc(nbFeatures, sizeof(double)); 
  plhs[0] = mxCreateDoubleMatrix(nbCentroids, nbFeatures, mxREAL);
  centroids = mxGetPr(plhs[0]);

  twoSigmaSquared = 2 * sigma * sigma;
  
  for(i=0;i<nbCentroids;i++){
 // mexPrintf("i %d\n", i);
    nbPtsInSphere = counts[i]; // there is sometimes no points in there
    neighborCell = mxGetCell(prhs[1], i); // mxArray for cell of neighborIndices
    neighbors = mxGetPr(neighborCell);
    distanceCell = mxGetCell(prhs[2], i); // mxArray for cell of neighborIndices
    distances = mxGetPr(distanceCell);
    totalWeight = 0; // init
    for(j=0;j<nbFeatures;j++){ // init coords to zero
      weightTimesCentroid[j] = 0.0;
    }
    for(k=0;k<nbPtsInSphere;k++){
 //   mexPrintf("k %d\n", k);
      neighborIndex = neighbors[k] -1; // - 1 as arrays here are zero-based
      neighborDistance = distances[k];
      neighborWeight = weights[neighborIndex] * exp(-(neighborDistance * neighborDistance)/twoSigmaSquared);
      totalWeight += neighborWeight;
      for(j=0;j<nbFeatures;j++){      
		 //	mexPrintf("j %d\n", j);
			vectCoordIndex = j * nbFeatureVects + neighborIndex;
		 //	mexPrintf("vectCoordIndex %d\n", vectCoordIndex);
			vectCoord = *(featureVects + vectCoordIndex);
		 //	mexPrintf("vectCoord %f\n", vectCoord);
			weightTimesCentroid[j] += neighborWeight * vectCoord;
      }
    }
    for(j=0;j<nbFeatures;j++){
      centroidCoordIndex = j * nbCentroids + i;
      centroidCoordPtr = centroids + centroidCoordIndex;
      if(totalWeight == 0){
		oldCentroidCoordPtr = oldCentroids + centroidCoordIndex;
		*centroidCoordPtr = *oldCentroidCoordPtr;
      }
      else{
 //	mexPrintf("% totalWeight %f\n", totalWeight);
		*centroidCoordPtr = weightTimesCentroid[j]/totalWeight;
      }
    }
  }

  mxFree(weightTimesCentroid);
}

/*******************************************************************************/
