function [clusterCenters, clusterTrace, closeVectIndexList, ranges] = ...
			meanShiftOptimalTrace(featureList, dataFlags, importances, doGaussian)
%
% Usage: [clusterCenters, clusterTrace, closeVectIndexList] = ...
%         meanShifOptimalTrace(featureList, dataFlags, importances, doGaussian)
%
% Combination of meanShiftMultiPass and findOptimalClustering
% % Return a trace in which each region corresponds to the longest living clusters when radius is increased
% and the corresponding clusters

% Check if there was a grouping between parent trace and child trace
% if there is, relabel new cluster with a new global label

% We collect traces at each clustering step to build a tree of coarse-to-fine segmentation
% Also, we collect number of regions at each step to be able to trace a curve nbRegions = f(radius)

% This is a general mean shift calculation with no assumption on the meaning of the features
% We are using fastMergeDups, a merging of duplicates based on sorting 
% which is faster than the one using range_search for data with a lot of preexisting clusters
%
% clusterTrace is the list of indices showing which cluster each feature list item belongs to
% Therefore length(clusterTrace) = length(featureList)
% and max(clusterTrace) = length(clusterCenters)
%
% clusterCenters is a matrix N x n, with N number of clusters and rows showing their components
%
% dataFlags show which feature vectors can be used (here some don't have angle components
% In this version we group neighbors only once
%
% Functions used by meanShiftEngine3:
%
%   nn_prepare, range_search from TSTools
%	keepNonZeroFeatures
%   normalizeFeatures
%   fastMergeDups
%   mergeTraces
%	findTreeShiftedVectors2
%   shiftTreeFeatures
%   mergeShiftFlags
%	findWeights
%   groupTreeNeighbors
%   findTreeNeighbors 
%   inverseMapFeatures
%

% deltaRadius = radius/nbSteps; % we increment radius by deltaRadius at every step
% We multiply initial radius by 2 at every iteration step

% In this data-collecting version, keep traces
disp('Build nonzero feature list and trace');
[nonZeroFeatureList, nonZeroTrace] = keepNonZeroFeatures(featureList, dataFlags);

disp('Map ranges of each component to range [0,1] (map to hypercube of size 1)');
% It is better to normalize after removing the zeros, for which data are unknown
[normalizedFeatures, ranges] = normalizeFeatures(double(nonZeroFeatureList), importances);

disp('Find weighted background of features');

quantizError = 1/100;
[weightedFeatures, featureWeights, tempTrace] = fastMergeDups(normalizedFeatures, quantizError);
  
weightedTrace = mergeTraces(nonZeroTrace, tempTrace);

% radius = 0.5;
radius = 0.78;
% nbSteps = 20;
nbSteps = 18;
% nbSteps = 16;
% factor = 2;
% factor = 1.5;
factor = 1.25;
% factor = 1.1;

initRadius = radius/(factor^(nbSteps-1)); % to get radius at end of loop
clusters = weightedFeatures;
weights = featureWeights;
trace = weightedTrace;
iteratedRadius = initRadius;
initNbClusters = 1000; % nb of clusters for starting region age computation
% initNbClusters = 400; % nb of clusters for starting region age computation
minNbClusters = 10; 

for i = 1:nbSteps
    disp(' ');
    disp(' '); % blank lines
    disp(['Iterated radius at hierarchy step ', num2str(i),' is ', num2str(iteratedRadius)]);
    quantizError = iteratedRadius/20;
    treeData = nn_prepare(clusters);

    [newClusters, newTrace] = findTreeShiftedVectors(clusters, weights, treeData, iteratedRadius, quantizError, doGaussian);
    trace = mergeTraces(trace, newTrace);
    newWeights = findWeights(trace);
    
    if i == 1 % first pass through loop
        firstTreeData = treeData; % treeData for finding nearest neigbors at end of code
    end
    
    nbClusters = max(trace);
    disp(['\\\\\ Nb of clusters ', num2str(nbClusters),' /////']);
%     if nbClusters < minNbClusters
%         break;
%     end; % we don't want too few regions
    if (nbClusters<=initNbClusters) % start scale-space clustering only once we have less than 1000 regions
		if 0
			disp(' Group neighbors '); % try without this step too to speed things up
			[newClusters, localTrace] = findMaxNeighbors(newClusters, newWeights, iteratedRadius/2); 
			trace = mergeTraces(trace, localTrace);
		end;
%       [bestClusters, bestTrace, bestAges] = keepOldestRegions2(newClusters, trace);
%        [bestClusters, bestTrace, bestAges] = keepOldestRegions(newClusters, trace);
         disp('%%%%%%%%%%%%%%%%%%%%%%% New scale-space clustering using mean children ages %%%%%%%%%%%%%%%%%%%%%%');
         [bestClusters, bestTrace, bestAges] = keepOldestMeanAgeRegions(newClusters, trace);
   end % if < initNbClusters

    clusters = newClusters; % so that when we exit, clusters and newClusters remain different
    weights = newWeights;
    iteratedRadius = factor * iteratedRadius;
end; % for i

% munlock keepOldestRegions2;
%clear keepOldestRegions2;
munlock keepOldestMeanAgeRegions; % Needed to clear persistent variables stuck in keepOldestMeanAgeRegions function
clear keepOldestMeanAgeRegions;

disp('Reassign bestTrace labels');
% Labels in bestTrace have missing elements in label sequence;
% replace current labels using a continuous sequence from 1 to nb of labels

[bestClusters2, bestTrace2] = reassignLabels(bestClusters, bestTrace);

disp(['@@@@@@@ Oldest region age is ',num2str(max(bestAges)), ' @@@@@@@']);
disp(['@@@@@@@ Final nb of clusters is ',num2str(length(bestClusters)), ' @@@@@@@']);


disp('Find vector of featureList that is closest to each clusterCenter');
% Use color of nearest neighbor featureList vector as color of cluster
% This should be a cleaner color than average color

closeVectIndexList = findCloseVects(weightedFeatures, weightedTrace, bestClusters2, firstTreeData);

clusterTrace = bestTrace2;
disp('Denormalize clusters');
clusterCenters = inverseMapFeatures(bestClusters2, ranges, importances);




