function clipTubes = findAffineFlowClipTubes(clipDir, scaledDir, motionXDir, motionYDir, vDefinedDir, ...
						sequenceName, startFrame, endFrame, showPictures)
%
% Find tubes with longest lives over the hierarchical mean shift tree
% In this version, we read precomputed optical flow found in subdirectories MotionX & MotionY
% and we read the reduced scales RGB images found in subdirectory QuarterScale

% 1. Read all frames and put in video stack
disp('Read stack of color images and flow images');

[rgb3D, angleMapX, angleMapY, vDefined] = readMapImages(clipDir, scaledDir, motionXDir, motionYDir, vDefinedDir, ...
							 sequenceName, startFrame, endFrame);

[nbFrames, nbRows, nbCols] = size(angleMapX); % these are trimmed dimensions
nbPixels = nbFrames * nbRows * nbCols;

nbUndefinedPixels = nbPixels - sum(vDefined(:));
if nbUndefinedPixels > 0.9 * nbPixels % do not process; sequence is probably black
    clipTubes = [];
    return
end

% 5. Make LUV images
disp('Make LUV image stack');
% [rImage, gImage] = getRGImages3D(rgb3D);
[LImage,uImage,vImage] = getLuvImages3D(rgb3D);
[frameCountBlock, rowCountBlock, colCountBlock] = makeCountBlocks(nbFrames, nbRows, nbCols);

% 6. Make feature vector list
% Note: Here we are not using the normal distances, only the x and y positions. Less noise since independent of optical flow
disp('Make list of feature vectors');

featureList = [angleMapX(:), angleMapY(:), colCountBlock(:), rowCountBlock(:), LImage(:), uImage(:), vImage(:)]; % 7 components

dataFlags = vDefined(:);


% 8. Run MeanShiftEngine

disp('Start Mean Shift Engine'); 

time1 = cputime;

importances = [0.8,0.8,0.6,0.6,1,1,1];
doGaussian = 1;

[clusterCenters, clusterTrace, closeVectIndexList, ranges] = meanShiftOptimalTrace(featureList, dataFlags, importances, doGaussian); 
% minNbPix = round(nbRows * nbCols/400);
% minNbPix = 10;
minNbPix = 4;
minAngle = 2; % discard clusters moving less than 4 degrees away from vertical
% Keep only larger clusters and discard cluster at angles around 90 degrees
% [clipTubes, bestTrace] = selectBestClusters(clusterCenters, clusterTrace, nbFrames, minNbPix);
clusterColors = findClusterColors3D2(rgb3D, closeVectIndexList);
clusterTrace3D = reshape(clusterTrace, nbFrames, nbRows, nbCols);

disp('Find line segments along cluster lines');
clusterSegments =  findClusterSegments3D(clusterCenters, clusterTrace3D); % Note: not quite correct since we don't use projDists. Fix

[clipTubes, bestTrace, bestIndexList] = selectBestMovingClusters(clusterCenters, clusterTrace, nbFrames, minNbPix, minAngle);

disp('Save workspace as a *.mat file');
save affineFlowClipTubeResults;

if ~isempty(clipTubes)  % coloring bombs if single label 
	bestClusterColors = clusterColors(bestIndexList,:);
	bestClusterSegments = clusterSegments(bestIndexList,:);
	% Now use plot routines to show the lines corresponding to moving objects
	% Draw these lines with the right colors in perspective projections
	% using intersection between lines and cube boundaries
	[lineList, colorList, thicknessList] = listClusterLines3D(nbFrames, nbRows, nbCols, clipTubes, bestClusterSegments, bestClusterColors);
	draw3DLines(nbFrames, nbRows, nbCols, lineList, colorList, thicknessList);
end

if showPictures
    bestTrace3D = reshape(bestTrace, nbFrames, nbRows, nbCols);
    mask3D = ~logical(bestTrace3D == 0);
    mask4D = repmat(mask3D,[1,1,1,3]);

    % Find cluster colors

    % Find segment labels
    % Make clusterTrace a 3D stack giving cluster label for each pixel of stack
%     disp('Label pixels by reshaping cluster trace to video stack shape');

    coloredClusterTrace = colorEachLabel(clusterTrace3D);
%     coloredClusterTrace = colorEachLabel3(clusterTrace3D);
    slicePixelBlock(coloredClusterTrace);
    drawnow;
    % Segment shot stack by giving pixels the values of the clusters which they clustered to
    disp('Find color and angle segmentations');
    [segmented3D, angle3DX, angle3DY] = segmentShotsFromClusters3D(clusterCenters, clusterColors, clusterTrace3D);
    segmentedBorders = colorsFromClusters3D(clusterColors, clusterTrace3D);
    slicePixelBlock(segmented3D);
    slicePixelBlock(segmentedBorders);
    drawnow;
%   slicePixelBlock(angle3DX);
    % slicePixelBlock(angle3DY);
    if ~isempty(clipTubes)  % coloring bombs if single label 
        coloredBestTrace = colorEachLabel(bestTrace3D);
        slicePixelBlock(coloredBestTrace);
	%     angleMask = ~logical(angle3DX>85 & angle3DX<95 & angle3DY>85 & angle3DY<95);
        slicePixelBlock(uint8(double(angle3DX) .* double(mask3D)));
        maskedRGB3D = double(rgb3D) .* double(mask4D);
        slicePixelBlock(uint8(maskedRGB3D));
        drawnow;
    end
% Now use plot routines to show the lines corresponding to moving objects
% Draw these lines with the right colors in perspective projections
% using intersection between lines and cube boundaries
% 	[lineList, colorList, thicknessList] = listClusterLines3D(nbFrames, nbRows, nbCols, bestClusterCenters, bestClusterSegments, bestClusterColors);
% 	draw3DLines(nbFrames, nbRows, nbCols, lineList, colorList, thicknessList);
  
    
end
